"""
datazen - A base class to be extended for runtime data loading and storing.
"""

# built-in
from collections import defaultdict
import logging
from typing import List, NamedTuple
import threading

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.environment import EnvironmentNamespace, clone

LOG = logging.getLogger(__name__)
SLUG_DELIM = "-"


class Task(NamedTuple):
    """Parameters identifying a task."""

    variant: str
    name: str

    @property
    def slug(self) -> str:
        """Convert this task into its 'slug' form."""

        return f"{self.variant}{SLUG_DELIM}{self.name}"


class TaskResult(NamedTuple):
    """
    Return value for a task, express whether or not the task succeeded and if
    this tasks' result should be considered 'new' from the last time it was
    evaluated.
    """

    success: bool
    fresh: bool


def dep_slug_unwrap(slug: str, default_op: str) -> Task:
    """
    From a slug String, determine the operation + target pair, if the
    operation isn't specified use a default.
    """

    result = Task(default_op, slug)
    if SLUG_DELIM in slug:
        split = slug.split(SLUG_DELIM, 1)
        result = Task(split[0], split[1])
    return result


class BaseEnvironment:
    """The base class for environment loading-and-storing management."""

    def __init__(self, default_ns: str = ROOT_NAMESPACE):
        """
        Manage environments by names, set up a dictionary with a root
        namespace.
        """

        self.namespaces = {}
        self.namespaces[default_ns] = EnvironmentNamespace(default_ns)
        self.lock = threading.RLock()

    def add_namespace(self, name: str, clone_root: bool = True) -> None:
        """Add a new namespace, optionally clone from the existing root."""

        with self.lock:
            self.namespaces[name] = EnvironmentNamespace(name)
            if clone_root and name != ROOT_NAMESPACE:
                clone(self.namespaces[ROOT_NAMESPACE], self.namespaces[name])

    def get_valid(self, name: str = ROOT_NAMESPACE) -> bool:
        """Get the 'valid' flag for a namespace."""

        return self.namespaces[name].valid

    def set_valid(self, value: bool, name: str = ROOT_NAMESPACE) -> None:
        """Set the 'valid' flag for a namespace."""

        self.namespaces[name].valid = value

    def get_to_load(
        self, dir_type: DataType, name: str = ROOT_NAMESPACE
    ) -> List[str]:
        """Proxy for a namespace's get_to_load."""

        return self.namespaces[name].get_to_load(dir_type)

    def get_data(self, dir_type: DataType, name: str = ROOT_NAMESPACE) -> dict:
        """Get the raw data for a directory type from a namespace."""

        return self.namespaces[name].data[dir_type]

    def unload_all(self, name: str = ROOT_NAMESPACE) -> None:
        """Unload all of the directories for a namespace."""

        return self.namespaces[name].unload_all()

    def update_load_state(
        self,
        dir_type: DataType,
        to_load: List[str],
        name: str = ROOT_NAMESPACE,
    ) -> int:
        """Proxy for update_load_state for a namespace."""

        return self.namespaces[name].update_load_state(dir_type, to_load)

    def add_dir(
        self,
        dir_type: DataType,
        dir_path: str,
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> bool:
        """Proxy for add_dir for a namespace."""

        return self.namespaces[name].add_dir(
            dir_type, dir_path, rel_path, allow_dup
        )

    def add_dirs(
        self,
        dir_type: DataType,
        dir_paths: List[str],
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> int:
        """Proxy for add_dirs for a namespace."""

        return self.namespaces[name].add_dirs(
            dir_type, dir_paths, rel_path, allow_dup
        )

    def get_namespace(
        self, key_name: str, target: str, target_data: dict
    ) -> str:
        """
        Determine the namespace that a target should use, in general they
        all should be unique unless they don't load anything new.
        """

        load_deps = defaultdict(list)
        schema_deps = ["schemas", "schema_types"]
        load_deps["compiles"] = ["configs", "variables"] + schema_deps
        load_deps["renders"] = ["templates"]

        # add a unique namespace for this target if it loads
        # any new data as to not load any of this data upstream,
        # but still cache it (edge cases here?)
        load_dep_list = load_deps[key_name]
        for load_dep in load_dep_list:
            if load_dep in target_data:
                namespace = Task(key_name, target).slug
                self.add_namespace(namespace)
                return namespace

        # don't make a new namespace if we don't load
        # new data
        return ROOT_NAMESPACE



"""
datazen - An environment extension that exposes command-line command execution.
"""

# built-in
from collections import defaultdict
import logging
import os
import subprocess
from typing import List

# internal
from datazen.environment.base import TaskResult
from datazen.environment.task import TaskEnvironment, get_path


class CommandEnvironment(TaskEnvironment):
    """Exposes command-line commanding capability to the environment."""

    def __init__(self):
        """Add the 'commands' handle."""

        super().__init__()
        self.handles["commands"] = self.valid_command

    def valid_command(
        self,
        entry: dict,
        _: str,
        __: dict = None,
        deps_changed: List[str] = None,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> TaskResult:
        """Perform the command specified by the entry."""

        cmd = [entry["command"]]
        if "arguments" in entry and entry["arguments"]:
            cmd += entry["arguments"]

        task_data = self.task_data["commands"][entry["name"]]

        # determine if the command needs to run
        file_exists = True
        if "file" in entry:
            file_exists = os.path.isfile(get_path(entry, "file"))
        force = "force" in entry and entry["force"]
        if not force and (
            not deps_changed and file_exists and entry["name"] in task_data
        ):
            return TaskResult(True, False)

        result = subprocess.run(cmd, capture_output=True)

        task_data[entry["name"]] = defaultdict(str)
        data = task_data[entry["name"]]
        data["args"] = result.args
        data["stdout"] = result.stdout.decode()
        data["stderr"] = result.stderr.decode()
        data["returncode"] = str(result.returncode)

        # log information about failures
        if result.returncode != 0:
            logger.error("command '%s' failed!", entry["command"])
            logger.error("args: %s", ", ".join(result.args))
            logger.error("exit: %d", result.returncode)
            logger.error("stdout:")
            print(data["stdout"])
            logger.error("stderr:")
            print(data["stderr"])

        return TaskResult(result.returncode == 0, True)



from datazen import DESCRIPTION as DESCRIPTION, VERSION as VERSION
from datazen.app import add_app_args as add_app_args, entry as entry
from typing import List

def main(argv: List[str] = ...) -> int: ...



from datazen.enums import DataType as DataType
from datazen.paths import resolve_dir as resolve_dir
from typing import Any, List

class EnvironmentNamespace:
    directories: Any
    data: Any
    data_clone_strategy: Any
    valid: bool
    lock: Any
    logger: Any
    def __init__(self, name: str) -> None: ...
    def get_to_load(self, dir_type: DataType) -> List[str]: ...
    def unload(self, dir_type: DataType) -> None: ...
    def unload_all(self) -> None: ...
    def update_load_state(self, dir_type: DataType, to_load: List[str]) -> int: ...
    def add_dir(self, dir_type: DataType, dir_path: str, rel_path: str = ..., allow_dup: bool = ...) -> bool: ...
    def add_dirs(self, dir_type: DataType, dir_paths: List[str], rel_path: str = ..., allow_dup: bool = ...) -> int: ...

def clone(env: EnvironmentNamespace, update: EnvironmentNamespace) -> EnvironmentNamespace: ...



"""
datazen - A target-type for grouping other target-tasks together.
"""

# built-in
import logging
from typing import List

# internal
from datazen.environment.base import TaskResult
from datazen.environment.task import TaskEnvironment


class GroupEnvironment(TaskEnvironment):
    """Leverages a task-environment to group tasks together."""

    def __init__(self):
        """Add the 'renders' handle."""

        super().__init__()
        self.handles["groups"] = self.valid_group

    def valid_group(
        self,
        entry: dict,
        _: str,
        dep_data: dict = None,
        deps_changed: List[str] = None,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> TaskResult:
        """Stub task to group other tasks."""

        if dep_data is not None:
            self.task_data["groups"][entry["name"]] = dep_data
        changed = bool(deps_changed)
        if changed:
            logger.info("group '%s' updated", entry["name"])
        return TaskResult(True, changed)



import logging
from cerberus import Validator
from datazen.classes.valid_dict import ValidDict as ValidDict
from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles, load_dir as load_dir
from typing import Any, Dict, Iterator, List

LOG: Any

def load(directories: List[str], require_all: bool = ..., loads: LoadedFiles = ...) -> dict: ...
def add_global_schemas(schema_data: Dict[str, dict], logger: logging.Logger = ...) -> None: ...
def remove_global_schemas(schema_data: Dict[str, dict], logger: logging.Logger = ...) -> None: ...
def load_types(directories: List[str], loads: LoadedFiles = ...) -> Dict[str, dict]: ...
def validate(schema_data: Dict[str, Validator], data: dict, logger: logging.Logger = ...) -> bool: ...
def inject_custom_schemas(schema_data: Dict[str, dict], should_inject: bool = ..., logger: logging.Logger = ...) -> Iterator[None]: ...



import logging
from cerberus import Validator as Validator
from collections import UserDict
from typing import Any

class ValidDict(UserDict):
    name: Any
    validator: Any
    valid: Any
    logger: Any
    def __init__(self, name: str, data: dict, schema: Validator, logger: logging.Logger = ...) -> None: ...



"""
datazen - A data structure for runtime data.
"""

# built-in
from copy import copy, deepcopy
import logging
import os
import threading
from typing import Dict, List

# internal
from datazen.paths import resolve_dir
from datazen.enums import DataType


class EnvironmentNamespace:
    """The base class for management of the environment data structure."""

    def __init__(self, name: str):
        """Base constructor."""

        self.directories: Dict[DataType, List[dict]] = {}
        for dtype in DataType:
            self.directories[dtype] = []

        self.data: Dict[DataType, dict] = {}
        for dtype in DataType:
            self.data[dtype] = {}

        self.data_clone_strategy: Dict[DataType, dict] = {
            DataType.CONFIG: {"should": True, "method": deepcopy},
            DataType.SCHEMA: {"should": True, "method": deepcopy},
            DataType.SCHEMA_TYPES: {"should": False},
            DataType.VARIABLE: {"should": True, "method": deepcopy},
            DataType.TEMPLATE: {"should": True, "method": copy},
        }

        self.valid = True
        self.lock = threading.Lock()
        self.logger = logging.getLogger(f"namespace.{name}")

    def get_to_load(self, dir_type: DataType) -> List[str]:
        """
        Build a list of the yet-to-be-loaded directories for a given data
        type.
        """

        with self.lock:
            dir_data = self.directories[dir_type]

            to_load = []
            for dir_inst in dir_data:
                if not dir_inst["loaded"]:
                    to_load.append(dir_inst["path"])

        return to_load

    def unload(self, dir_type: DataType) -> None:
        """Mark all directories for a given type as un-loaded."""

        with self.lock:
            for dir_inst in self.directories[dir_type]:
                dir_inst["loaded"] = False
            self.data[dir_type] = {}

    def unload_all(self) -> None:
        """Mark all directories as unloaded."""

        for key in self.data:
            self.unload(key)

    def update_load_state(self, dir_type: DataType, to_load: List[str]) -> int:
        """
        Update the load states of directories in 'to_load' for a given
        data type.
        """

        with self.lock:
            dir_data = self.directories[dir_type]
            loaded = 0

            for dir_inst in dir_data:
                if dir_inst["path"] in to_load:
                    self.logger.debug(
                        "loaded '%s' as '%s'", dir_inst["path"], dir_type.value
                    )
                    dir_inst["loaded"] = True
                    loaded = loaded + 1

        self.logger.debug(
            "loaded %d new directories for '%s'", loaded, dir_type.value
        )

        return loaded

    def add_dir(
        self,
        dir_type: DataType,
        dir_path: str,
        rel_path: str = ".",
        allow_dup: bool = False,
    ) -> bool:
        """Add a directory to be loaded for a given data type."""

        with self.lock:
            dir_list = self.directories[dir_type]
            dir_path = resolve_dir(dir_path, rel_path)

            for dir_data in dir_list:
                if dir_path == dir_data["path"]:

                    def noop(*_, **__):
                        pass

                    log_fn = noop if allow_dup else noop
                    log_fn(
                        "not adding duplicate directory '%s' to '%s'",
                        dir_path,
                        dir_type.value,
                    )
                    return allow_dup

            if not os.path.isdir(dir_path):
                self.logger.error(
                    "'%s' isn't a directory, not adding to '%s'",
                    dir_path,
                    dir_type.value,
                )
                return False

            dir_list.append({"path": dir_path, "loaded": False})

        self.logger.debug("added '%s' to '%s'", dir_path, dir_type.value)
        return True

    def add_dirs(
        self,
        dir_type: DataType,
        dir_paths: List[str],
        rel_path: str = ".",
        allow_dup: bool = False,
    ) -> int:
        """
        Add multiple directories for a given data type, return the number of
        directories added.
        """

        dirs_added = 0

        for dir_path in dir_paths:
            if self.add_dir(dir_type, dir_path, rel_path, allow_dup):
                dirs_added = dirs_added + 1

        if dirs_added != len(dir_paths):
            self.logger.error(
                "only loaded %d / %d directories, marking invalid",
                dirs_added,
                len(dir_paths),
            )
            self.valid = False

        return dirs_added


def clone(
    env: EnvironmentNamespace, update: EnvironmentNamespace
) -> EnvironmentNamespace:
    """Create a clone (deep copy) of an existing Environment."""

    with env.lock:
        # all we need to do is copy all of the attributes
        update.directories = deepcopy(env.directories)

        # clone data based on the source's strategy
        for dtype, strat in env.data_clone_strategy.items():
            if strat["should"]:
                update.data[dtype] = strat["method"](env.data[dtype])

        update.valid = env.valid

    return update



from typing import Any

PKG_NAME: str
VERSION: str
DESCRIPTION: str
DEFAULT_TYPE: str
DEFAULT_MANIFEST: Any
DEFAULT_DIR: Any
CACHE_SUFFIX: str
GLOBAL_KEY: str

def to_private(input_str: str) -> str: ...

ROOT_NAMESPACE: Any



import jinja2
from datazen import ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.enums import DataType as DataType
from datazen.environment.base import BaseEnvironment as BaseEnvironment
from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles
from typing import Dict, List

class TemplateEnvironment(BaseEnvironment):
    def load_templates(self, template_loads: LoadedFiles = ..., name: str = ...) -> Dict[str, jinja2.Template]: ...
    def add_template_dirs(self, dir_paths: List[str], rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> int: ...



"""
datazen - A child class for adding variable-data loading capabilities to the
          environment dataset.
"""

# built-in
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.environment.base import BaseEnvironment
from datazen.load import LoadedFiles, DEFAULT_LOADS
from datazen.variables import load as load_variables


class VariableEnvironment(BaseEnvironment):
    """
    The variable-data loading environment mixin. Only requires the base
    environment capability to function.
    """

    def load_variables(
        self,
        var_loads: LoadedFiles = DEFAULT_LOADS,
        name: str = ROOT_NAMESPACE,
    ) -> dict:
        """Load variable data, resolve any un-loaded variable directories."""

        # determine directories that need to be loaded
        data_type = DataType.VARIABLE

        with self.lock:
            to_load = self.get_to_load(data_type, name)

            # load new data
            variable_data = self.get_data(data_type, name)
            if to_load:
                variable_data.update(load_variables(to_load, var_loads))
                self.update_load_state(data_type, to_load, name)

        return variable_data

    def add_variable_dirs(
        self,
        dir_paths: List[str],
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> int:
        """
        Add variable-data directories, return the number of directories added.
        """

        return self.add_dirs(
            DataType.VARIABLE, dir_paths, rel_path, name, allow_dup
        )



"""
datazen - A dict wrapper that enables simpler schema validation.
"""

# built-in
import logging
from collections import UserDict

# third-party
from cerberus import Validator


class ValidDict(UserDict):
    """
    An object that behaves like a dictionary but can have a provided schema
    enforced.
    """

    def __init__(
        self,
        name: str,
        data: dict,
        schema: Validator,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        """Initialize a named, ValidDict."""

        # initialize the dict
        if not isinstance(data, dict):
            data = {}
        super().__init__(data)

        self.name = name
        self.validator = schema
        self.valid = self.validator.validate(self.data)
        self.logger = logger
        if not self.valid:
            self.logger.error(
                "validation error(s) for '%s': %s",
                self.name,
                self.validator.errors,
            )
            self.logger.error("data: %s", data)
            self.logger.error("schema: %s", schema.schema)



"""
datazen - A centralized store for runtime data.
"""

# built-in
from collections import defaultdict
import logging
import os
from typing import List, Tuple

# internal
from datazen import CACHE_SUFFIX
from datazen.environment.base import dep_slug_unwrap
from datazen.environment.command import CommandEnvironment
from datazen.environment.group import GroupEnvironment
from datazen.environment.compile import CompileEnvironment
from datazen.environment.render import RenderEnvironment

LOG = logging.getLogger(__name__)


class Environment(
    CompileEnvironment, RenderEnvironment, GroupEnvironment, CommandEnvironment
):
    """A wrapper for inheriting all environment-loading capabilities."""

    def __init__(self, logger: logging.Logger = LOG):
        """Add a notion of 'visited' targets to the environment data."""

        super().__init__()
        self.visited = defaultdict(bool)
        self.default = "compiles"
        self.logger = logger

    def execute(
        self, target: str = "", should_cache: bool = True
    ) -> Tuple[bool, bool]:
        """Execute an arbitrary target."""

        # resolve a default target if one wasn't provided
        data = self.manifest["data"]
        if not target:
            if "default_target" in data:
                target = data["default_target"]
                self.logger.info("using default target '%s'", target)
            elif self.default in data and data[self.default]:
                target = data[self.default][0]["name"]
                self.logger.info("resolving first target '%s'", target)

        slug = dep_slug_unwrap(target, self.default)
        return self.handle_task(slug[0], slug[1], should_cache=should_cache)

    def execute_targets(self, targets: List[str]) -> bool:
        """
        Execute a list of named targets and return whether or not the entire
        set was successful.
        """

        # use a resolved, default target if none are specified
        if not targets and self.default_target():
            targets.append(self.default_target())

        for target in targets:
            task_result = self.execute(target, False)
            if not task_result[0]:
                self.logger.error("target '%s' failed", target)
                return False
            if not task_result[1]:
                self.logger.info("'%s' already satisfied", target)

        # write the cache at the end, if we were totally successful
        self.write_cache()
        return True

    def group(self, target: str) -> Tuple[bool, bool]:
        """Attempt to satisfy a 'group' target."""

        return self.handle_task("groups", target)

    def compile(self, target: str) -> Tuple[bool, bool]:
        """Execute a named 'compile' target from the manifest."""

        return self.handle_task("compiles", target)

    def render(self, target: str) -> Tuple[bool, bool]:
        """Execute a named 'render' target from the manifest."""

        return self.handle_task("renders", target)

    def command(self, target: str) -> Tuple[bool, bool]:
        """Execute a named 'command' target from the manifest."""

        return self.handle_task("commands", target)


def from_manifest(
    manifest_path: str,
    data_cache_name: str = "task_data",
    logger: logging.Logger = LOG,
) -> Environment:
    """Load an environment object from a schema definition on disk."""

    env = Environment(logger)

    # load the manifest
    if not env.load_manifest_with_cache(manifest_path):
        logger.error("couldn't load manifest at '%s'", manifest_path)
    else:
        data_cache = f".{data_cache_name}{CACHE_SUFFIX}"
        path = os.path.join(os.path.dirname(env.cache.cache_dir), data_cache)
        env.init_cache(os.path.abspath(path))

    return env



"""
datazen - APIs for loading data from directory trees.
"""

# built-in
from collections import defaultdict
from contextlib import contextmanager
import logging
import os
from typing import Any, Dict, List, Tuple, NamedTuple, Optional, Iterator

# internal
from datazen import GLOBAL_KEY
from datazen.paths import (
    get_path_list,
    advance_dict_by_path,
    walk_with_excludes,
    get_file_name,
)
from datazen.parsing import load as load_raw_resolve
from datazen.parsing import set_file_hash

LOG = logging.getLogger(__name__)


class LoadedFiles(NamedTuple):
    """
    A collection of data for files loaded at runtime (or, a continuation of
    this information loaded from a cache).
    """

    files: Optional[List[str]] = None
    file_data: Optional[Dict[str, dict]] = None


DEFAULT_LOADS = LoadedFiles()


@contextmanager
def data_added(key: Any, value: Any, data: dict = None) -> Iterator[dict]:
    """Inject a key-value pair into a dictionary, in a context."""

    if data is None:
        data = {}

    data[key] = value
    try:
        yield data
    finally:
        del data[key]


def meld_and_resolve(
    full_path: str,
    existing_data: dict,
    variables: dict,
    globals_added: bool = False,
    expect_overwrite: bool = False,
    is_template: bool = True,
) -> bool:
    """
    Meld dictionary data from a file into an existing dictionary, assume
    existing data is a template and attempt to resolve variables.
    """

    variables_root: dict = variables

    # allow directory/.{file_type} to be equivalent to directory.{file_type}
    key = get_file_name(full_path)
    if not key:
        data_dict = existing_data
    else:
        if key not in existing_data:
            existing_data[key] = {}
        data_dict = existing_data[key]
        if key in variables:
            variables = variables[key]

    # meld the data
    global_add_success: bool = False
    if globals_added:
        if GLOBAL_KEY not in variables:
            variables[GLOBAL_KEY] = variables_root[GLOBAL_KEY]
            global_add_success = True

    _, loaded = load_raw_resolve(
        full_path, variables, data_dict, expect_overwrite, is_template
    )

    if global_add_success:
        del variables[GLOBAL_KEY]

    return loaded


def load_dir(
    path: str,
    existing_data: dict,
    variables: dict = None,
    loads: LoadedFiles = DEFAULT_LOADS,
    expect_overwrite: bool = False,
    are_templates: bool = True,
    logger: logging.Logger = LOG,
) -> dict:
    """Load a directory tree into a dictionary, optionally meld."""

    if variables is None:
        variables = {}

    root_abs = os.path.abspath(path)
    for root, _, files in walk_with_excludes(path):
        logger.debug("loading '%s'", root)

        path_list = get_path_list(root_abs, root)
        variable_data = advance_dict_by_path(path_list, variables)

        # expose data globally, if it was provided
        added_globals: bool = False
        if GLOBAL_KEY not in variable_data:
            variable_data[GLOBAL_KEY] = variables
            added_globals = True
        else:
            logger.info(
                "can't add 'global' data to '%s', key was already found",
                root,
            )

        # extend the provided list of files that were newly loaded, or at
        # least have new content
        new = load_files(
            files,
            root,
            (
                advance_dict_by_path(path_list, existing_data),
                variable_data,
                added_globals,
            ),
            loads.file_data,
            expect_overwrite,
            are_templates,
        )
        if loads.files is not None:
            loads.files.extend(new)

        if added_globals:
            del variable_data[GLOBAL_KEY]

    return existing_data


def load_dir_only(
    path: str,
    expect_overwrite: bool = False,
    are_templates: bool = True,
    logger: logging.Logger = LOG,
) -> dict:
    """
    A convenient wrapper for loading just directory data from a path without
    worrying about melding data, resolving variables, enforcing schemas, etc.
    """

    return load_dir(
        path,
        defaultdict(dict),
        None,
        DEFAULT_LOADS,
        expect_overwrite,
        are_templates,
        logger,
    )


def load_files(
    file_paths: List[str],
    root: str,
    meld_data: Tuple[dict, dict, bool],
    hashes: Dict[str, dict] = None,
    expect_overwrite: bool = False,
    are_templates: bool = True,
) -> List[str]:
    """
    Load files into a dictionary and return a list of the files that are
    new or had hash mismatches.
    """

    new_or_changed = []

    # load (or meld) data
    for name in file_paths:
        full_path = os.path.join(root, name)
        assert os.path.isabs(full_path)
        success = meld_and_resolve(
            full_path,
            meld_data[0],
            meld_data[1],
            meld_data[2],
            expect_overwrite,
            are_templates,
        )
        if success and hashes is not None:
            success = set_file_hash(hashes, full_path)
        if success:
            new_or_changed.append(full_path)

    return new_or_changed



import logging
from datazen import VERSION as VERSION
from datazen.compile import write_dir as write_dir
from datazen.load import LoadedFiles as LoadedFiles, load_dir_only as load_dir_only
from datazen.parsing import dedup_dict_lists as dedup_dict_lists, set_file_hash as set_file_hash
from typing import Any, Dict, List

LOG: Any
DATA_DEFAULT: Any

class FileInfoCache:
    data: Any
    removed_data: Any
    cache_dir: str
    logger: Any
    def __init__(self, cache_dir: str = ..., logger: logging.Logger = ...) -> None: ...
    def load(self, cache_dir: str) -> None: ...
    def get_hashes(self, sub_dir: str) -> Dict[str, dict]: ...
    def check_hit(self, sub_dir: str, path: str, also_cache: bool = ...) -> bool: ...
    def get_loaded(self, sub_dir: str) -> List[str]: ...
    def describe(self) -> None: ...
    def get_data(self, name: str) -> LoadedFiles: ...
    def clean(self) -> None: ...
    def write(self, out_type: str = ...) -> None: ...

def copy(cache: FileInfoCache) -> FileInfoCache: ...
def meld(cache_a: FileInfoCache, cache_b: FileInfoCache) -> None: ...
def time_str(time_s: float) -> str: ...
def cmp_loaded_count(cache_a: FileInfoCache, cache_b: FileInfoCache, name: str) -> int: ...
def cmp_loaded_count_from_set(cache_a: FileInfoCache, cache_b: FileInfoCache, name: str, files: List[str]) -> int: ...
def cmp_total_loaded(cache_a: FileInfoCache, cache_b: FileInfoCache, known_types: List[str], load_checks: Dict[str, List[str]] = ..., logger: logging.Logger = ...) -> int: ...
def sync_cache_data(cache_data: dict, removed_data: Dict[str, List[str]]) -> dict: ...
def remove_missing_hashed_files(data: dict, removed_data: Dict[str, List[str]]) -> dict: ...
def remove_missing_loaded_files(data: dict) -> dict: ...



import logging
from datazen import ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.enums import DataType as DataType
from datazen.environment.schema import SchemaEnvironment as SchemaEnvironment
from datazen.environment.variable import VariableEnvironment as VariableEnvironment
from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles
from typing import List

class ConfigEnvironment(VariableEnvironment, SchemaEnvironment):
    configs_valid: bool
    def __init__(self) -> None: ...
    def load_configs(self, cfg_loads: LoadedFiles = ..., var_loads: LoadedFiles = ..., sch_loads: LoadedFiles = ..., sch_types_loads: LoadedFiles = ..., name: str = ..., logger: logging.Logger = ...) -> dict: ...
    def add_config_dirs(self, dir_paths: List[str], rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> int: ...



from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles, load_dir as load_dir
from typing import List

def load(directories: List[str], loads: LoadedFiles = ...) -> dict: ...



import logging
from datazen.paths import get_file_ext as get_file_ext
from typing import Any, List, NamedTuple, TextIO

LOG: Any

class LoadResult(NamedTuple):
    data: dict
    success: bool

def get_json_data(data_file: TextIO, logger: logging.Logger = ...) -> LoadResult: ...
def get_yaml_data(data_file: TextIO, logger: logging.Logger = ...) -> LoadResult: ...
def load_stream(data_stream: TextIO, data_path: str, logger: logging.Logger = ...) -> LoadResult: ...
def dedup_dict_lists(data: dict) -> dict: ...
def merge_dicts(dicts: List[dict], expect_overwrite: bool = ..., logger: logging.Logger = ...) -> dict: ...
def merge(dict_a: dict, dict_b: dict, path: List[str] = ..., expect_overwrite: bool = ..., logger: logging.Logger = ...) -> dict: ...
def load(data_path: str, variables: dict, dict_to_update: dict, expect_overwrite: bool = ..., is_template: bool = ..., logger: logging.Logger = ...) -> LoadResult: ...
def get_hash(data: str, encoding: str = ...) -> str: ...
def get_file_hash(path: str) -> str: ...
def set_file_hash(hashes: dict, path: str, set_new: bool = ...) -> bool: ...



"""
datazen - APIs for working with file paths.
"""

# built-in
import os
import pkgutil
from typing import List, Iterator, Tuple

# third-party
import pkg_resources

# internal
from datazen import PKG_NAME

FMT_OPEN = "{"
FMT_CLOSE = "}"
EXCLUDES = [".git", ".svn", ".gitignore"]


def get_file_name(full_path: str) -> str:
    """From a full path to a file, get just the name of the file."""

    return ".".join(os.path.basename(full_path).split(".")[:-1])


def get_file_ext(full_path: str) -> str:
    """From a pull path to a file, get just the file's extension."""

    return os.path.basename(full_path).split(".")[-1]


def get_path_list(root_abs_path: str, current_abs_path: str) -> List[str]:
    """
    From a root directory and a child path, compute the list of directories,
    in order, mapping the root to the child.
    """

    assert len(current_abs_path) >= len(root_abs_path)
    assert root_abs_path in current_abs_path
    return current_abs_path[len(root_abs_path) :].split(os.sep)


def format_resolve_delims(
    value: str, fmt_data: dict, delim: str = ".", delim_replace: str = "_"
) -> str:
    """
    Attempt to resolve a format String with data, but handle replacements with
    custom delimeters correctly.
    """

    open_len = value.count(FMT_OPEN)
    assert open_len == value.count(FMT_CLOSE)

    # if no parameters are specified, just return the provided String
    if open_len == 0:
        return value

    # replace "a.b.c" with "a_b_c", fmt_data should be a flat Dict[str, str]
    new_data: dict = {}
    for key, item in fmt_data.items():
        assert isinstance(key, str)
        new_data[key.replace(delim, delim_replace)] = item

    # re-build the format String
    fstr = ""
    tmp_value = value
    for _ in range(open_len):
        start = tmp_value.index(FMT_OPEN) + 1
        end = tmp_value.index(FMT_CLOSE)
        fstr += tmp_value[: start - 1]
        fstr += "{"
        fstr += tmp_value[start:end].replace(delim, delim_replace)
        tmp_value = tmp_value[end + 1 :]
        fstr += "}"

    if len(fstr) < len(value):
        fstr += value[len(fstr) :]

    return fstr.format(**new_data)


def unflatten_dict(data: dict, delim: str = ".") -> dict:
    """
    Attempt to unflatten dictionary data based on String keys and a delimeter.
    """

    result: dict = {}

    for key, value in data.items():
        # unflatten any dictionaries in the value
        if isinstance(value, dict):
            value = unflatten_dict(value)
        elif isinstance(value, list):
            new_value = []
            for item in value:
                if isinstance(item, dict):
                    item = unflatten_dict(item)
                new_value.append(item)
            value = new_value

        # move data["a.b.c"] = value to data["a"]["b"]["c"] = value
        if isinstance(key, str):
            to_advance = key.split(delim)
            to_update = advance_dict_by_path(to_advance[:-1], result)
            to_update[to_advance[-1]] = value
        else:
            result[key] = value

    return result


def advance_dict_by_path(path_list: List[str], data: dict) -> dict:
    """
    Given a dictionary and a list of directory names, return the child
    dictionary advanced by each key, in order, from the provided data.
    """

    for path in path_list:
        if path and isinstance(data, dict):
            try:
                data = data[path]
            except KeyError:
                data[path] = {}
                data = data[path]

    return data


def get_package_data(relative_path: str) -> str:
    """Load a file from this package's data directory."""

    rel_path = os.path.join("data", relative_path)
    schema_raw = pkgutil.get_data(PKG_NAME, rel_path)
    schema_bytes = schema_raw if schema_raw else bytes()
    return schema_bytes.decode("utf-8")


def get_package_dir(relative_path: str) -> str:
    """Locate the path to a package-data directory."""

    return pkg_resources.resource_filename(
        __name__, os.path.join("data", relative_path)
    )


def resolve_dir(data: str, rel_base: str = "") -> str:
    """
    Turn directory data into an absolute path, optionally from a relative
    base.
    """

    data = data.replace("/", os.sep)
    data = data.replace("\\", os.sep)
    if not os.path.isabs(data) and rel_base:
        data = os.path.join(rel_base, data)
    return os.path.abspath(data)


def walk_with_excludes(
    path: str, excludes: List[str] = None
) -> Iterator[Tuple[str, List[str], List[str]]]:
    """
    Behaves like os.walk but attempts to skip iterations that would enter
    directories we want to exclude (and thus not traverse).
    """

    if excludes is None:
        excludes = EXCLUDES

    for root, dirnames, filenames in os.walk(path):
        dirs = root.split(os.sep)

        # don't yield any directories that have an excluded path
        if any(x in dirs for x in excludes):
            continue

        yield root, dirnames, filenames



"""
datazen - An environment extension that exposes rendering capabilities.
"""

# built-in
import logging
import os
from typing import Dict, List, Optional

# third-party
import jinja2

# internal
from datazen import GLOBAL_KEY, to_private
from datazen.environment.base import dep_slug_unwrap, TaskResult
from datazen.environment.task import TaskEnvironment, get_path
from datazen.fingerprinting import build_fingerprint
from datazen.load import data_added
from datazen.paths import get_file_ext
from datazen.targets import resolve_dep_data

LOG = logging.getLogger(__name__)


def render_name_to_key(name: str) -> str:
    """Convert the name of a render target with a valid dictionary key."""

    return name.replace(".", "_")


def indent_str(data: str, indent: int, sep: str = os.linesep) -> str:
    """
    Attempt to indent String data by some amount, based on some separator.
    """

    result = ""
    ind_str = " " * indent
    for line in data.split(sep):
        if line:
            result += ind_str + line + sep
        else:
            result += sep
    return result.rstrip()


def get_render_str(
    template: jinja2.Template,
    name: str,
    indent: int,
    data: dict = None,
    out_data: dict = None,
) -> str:
    """Render a template."""

    # add a self-reference key for convenience
    global_added: bool = False
    if data is not None and GLOBAL_KEY not in data:
        data[GLOBAL_KEY] = data
        global_added = True

    result = template.render(data).rstrip()

    if global_added:
        assert data is not None
        del data[GLOBAL_KEY]

    # add indents if requested
    result = indent_str(result, indent)

    if out_data is not None:
        out_data[render_name_to_key(name)] = result
    return result


def get_render_children(
    children: dict,
    dep_data: dict,
    default_op: str,
    indent: int,
    delimeter: str = "",
) -> None:
    """Build child dependency data."""

    result = []
    for child in children:
        slug = dep_slug_unwrap(child, default_op)
        assert slug[1] in dep_data
        assert isinstance(dep_data[slug[1]], str)
        result.append(dep_data[slug[1]])
    dep_data[to_private("children")] = indent_str(
        delimeter.join(result), indent
    )


class RenderEnvironment(TaskEnvironment):
    """Leverages a cache-equipped environment to render templates."""

    def __init__(self):
        """Add the 'renders' handle."""

        super().__init__()
        self.handles["renders"] = self.valid_render

    def perform_render(
        self,
        template: jinja2.Template,
        all_templates: Dict[str, jinja2.Template],
        path: Optional[str],
        entry: dict,
        data: dict = None,
        logger: logging.Logger = LOG,
    ) -> TaskResult:
        """
        Render a template to the requested path using the provided data.
        """

        try:
            out_data: dict = {}

            # add template objects to render context
            with data_added(
                to_private("templates"), all_templates, data
            ) as render_data:
                render_str = get_render_str(
                    template,
                    entry["name"],
                    entry["indent"],
                    render_data,
                    out_data,
                )

            # determine if the caller wanted a dynamic fingerprint or not,
            # if an indent is set, also disable it
            dynamic = True
            if (
                "no_dynamic_fingerprint" in entry
                and entry["no_dynamic_fingerprint"]
            ) or entry["indent"]:
                dynamic = False

            fprint = build_fingerprint(
                render_str, get_file_ext(get_path(entry)), dynamic=dynamic
            )

            # don't write a file, if requested
            if path is not None:
                with open(path, "w", encoding="utf-8") as render_out:
                    render_out.write(fprint + render_str + os.linesep)

            # save the output into a dict for consistency
            self.store_render(entry, out_data)
        except jinja2.exceptions.TemplateError as exc:
            logger.error(
                "couldn't render '%s' to '%s': %s", entry["name"], path, exc
            )
            return TaskResult(False, False)

        logger.info("(%s) rendered '%s'", entry["name"], path)

        return TaskResult(True, True)

    def store_render(self, entry: dict, data: dict) -> None:
        """Store data from the current render."""

        if "as" in entry and entry["as"]:
            name_key = render_name_to_key(entry["name"])
            data[entry["as"]] = data[name_key]
            if entry["as"] != name_key:
                del data[name_key]
        self.task_data["renders"][entry["name"]] = data

    def valid_render(
        self,
        entry: dict,
        namespace: str,
        dep_data: dict = None,
        deps_changed: List[str] = None,
        logger: logging.Logger = LOG,
    ) -> TaskResult:
        """Perform the render specified by the entry."""

        # determine the output that will be produced
        path = None
        if "no_file" not in entry or not entry["no_file"]:
            path = get_path(entry)

        if "indent" not in entry:
            entry["indent"] = 0

        # load templates
        templates = self.cached_load_templates(namespace)

        temp_name = entry["name"]
        if "key" in entry and entry["key"]:
            temp_name = entry["key"]

        if temp_name not in templates:
            logger.error(
                "no template for key '%s' found, options: %s",
                entry["name"],
                list(templates.keys()),
            )
            return TaskResult(False, False)

        template = templates[temp_name]

        # if dependencies aren't specified, use config data (but don't allow
        # an implicit 'compile')
        change_criteria = ["templates"]
        if not dep_data and "dependencies" not in entry:
            dep_data = self.cached_load_configs(namespace)
            change_criteria.append("configs")
            logger.debug(
                "no dependencies loaded for '%s', using config data",
                entry["name"],
            )

        # apply overrides if present
        if dep_data is not None:
            dep_data = resolve_dep_data(entry, dep_data)

        if "children" in entry:
            assert dep_data is not None
            get_render_children(
                entry["children"],
                dep_data,
                self.default,
                entry.get("child_indent", 0),
                entry.get("child_delimeter", ""),
            )

        # determine if we need to perform this render
        assert template.filename is not None
        load_checks = {"templates": [template.filename]}
        if self.already_satisfied(
            entry["name"], path, change_criteria, deps_changed, load_checks
        ):
            logger.debug("render '%s' satisfied, skipping", entry["name"])
            return TaskResult(True, False)

        return self.perform_render(template, templates, path, entry, dep_data)



import logging
from cerberus import Validator
from datazen import DEFAULT_DIR as DEFAULT_DIR, ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.classes.target_resolver import TargetResolver as TargetResolver
from datazen.environment.config import ConfigEnvironment as ConfigEnvironment
from datazen.environment.template import TemplateEnvironment as TemplateEnvironment
from datazen.parsing import load_stream as load_stream, merge_dicts as merge_dicts
from datazen.paths import get_package_data as get_package_data, get_package_dir as get_package_dir, resolve_dir as resolve_dir
from datazen.schemas import inject_custom_schemas as inject_custom_schemas, load_types as load_types
from typing import Any, List, Tuple

LOG: Any

def get_output_dir(data: dict, rel_path: str, default: str = ...) -> str: ...
def set_output_dir(data: dict, rel_path: str, default: str = ...) -> None: ...

class ManifestEnvironment(ConfigEnvironment, TemplateEnvironment):
    manifest: Any
    target_resolver: Any
    def __init__(self) -> None: ...
    def load_dirs(self, data: dict, rel_path: str, namespace: str = ..., allow_dup: bool = ...) -> None: ...
    def default_target(self) -> str: ...
    def load_manifest_reent(self, path: str, manifest_dir: str, params: dict, files: List[str], logger: logging.Logger = ...) -> Tuple[dict, bool]: ...
    def load_manifest(self, path: str = ..., logger: logging.Logger = ...) -> bool: ...

def get_manifest_schema(require_all: bool = ..., logger: logging.Logger = ...) -> Validator: ...
def validate_manifest(manifest: dict, logger: logging.Logger = ...) -> bool: ...



# =====================================
# generator=datazen
# version=1.8.0
# hash=599db18ed2c4013da36c4da6e132512a
# =====================================

"""
datazen - Useful defaults and other package metadata.
"""

PKG_NAME = "datazen"
VERSION = "1.8.1"
DESCRIPTION = "Compile and render schema-validated configuration data."

DEFAULT_TYPE = "yaml"
DEFAULT_MANIFEST = f"manifest.{DEFAULT_TYPE}"
DEFAULT_DIR = f"{PKG_NAME}-out"
CACHE_SUFFIX = "_cache"
GLOBAL_KEY = "global"


def to_private(input_str: str) -> str:
    """Convert some input to a 'private' String."""

    return f"__{input_str}__"


ROOT_NAMESPACE = to_private("root")



"""
datazen - This package's command-line entry-point application.
"""

# built-in
import argparse

# internal
from datazen.environment.integrated import from_manifest
from datazen import DEFAULT_MANIFEST


def entry(args: argparse.Namespace) -> int:
    """Execute the requested task."""

    result = 0
    env = from_manifest(args.manifest)
    if env.get_valid():
        # clean, if requested
        if args.sync:
            env.write_cache()
        if args.clean:
            env.clean_cache()
        elif args.describe:
            env.describe_cache()
        else:
            # execute targets
            result = int(not env.execute_targets(args.targets))
    else:
        result = 1
    return result


def add_app_args(parser: argparse.ArgumentParser) -> None:
    """Add application-specific arguments to the command-line parser."""

    parser.add_argument(
        "-m",
        "--manifest",
        default=DEFAULT_MANIFEST,
        help=("manifest to execute tasks from (default: " + "'%(default)s')"),
    )
    parser.add_argument(
        "-c",
        "--clean",
        action="store_true",
        help="clean the manifest's cache and exit",
    )
    parser.add_argument(
        "--sync",
        action="store_true",
        help=(
            "sync the manifest's cache (write-through) "
            + "with the state of the file system before "
            + "execution"
        ),
    )
    parser.add_argument(
        "-d",
        "--describe",
        action="store_true",
        help="describe the manifest's cache and exit",
    )
    parser.add_argument("targets", nargs="*", help="target(s) to execute")



import logging
from datazen.environment.base import TaskResult as TaskResult
from datazen.environment.task import TaskEnvironment as TaskEnvironment, get_path as get_path
from typing import List

class CommandEnvironment(TaskEnvironment):
    def __init__(self) -> None: ...
    def valid_command(self, entry: dict, _: str, __: dict = ..., deps_changed: List[str] = ..., logger: logging.Logger = ...) -> TaskResult: ...



from datazen import PKG_NAME as PKG_NAME, VERSION as VERSION
from datazen.parsing import get_hash as get_hash
from typing import List, Tuple

BARRIER: str

def get_comment_data(file_data: str, dynamic: bool = ...) -> List[Tuple[str, str]]: ...
def apply_barriers(comment_lines: List[str], char: str) -> None: ...
def resolve_encapsulation(comment_lines: List[str], file_ext: str) -> str: ...
def build_fingerprint(file_data: str, file_ext: str, char: str = ..., dynamic: bool = ...) -> str: ...



"""
datazen - An interface for turning a dictionary into various serialized forms.
"""

# built-in
import io
import json
import logging
import os
from typing import Tuple

# third-party
from ruamel.yaml import YAML

# internal
from datazen import DEFAULT_TYPE

LOG = logging.getLogger(__name__)


def write_dir(directory: str, data: dict, out_type: str = "json") -> None:
    """Write dictionary data to the file-system."""

    os.makedirs(directory, exist_ok=True)
    for key, val in data.items():
        key_path = os.path.join(directory, f"{key}.{out_type}")
        with open(key_path, "w", encoding="utf-8") as key_file:
            key_file.write(str_compile(val, out_type))


def str_compile(
    configs: dict, data_type: str, logger: logging.Logger = LOG
) -> str:
    """
    Serialize dictionary data into the String-form of a specific,
    serializeable type.
    """

    ostream = io.StringIO()

    # serialize the data
    if data_type == "json":
        result = json.dumps(configs, indent=4, sort_keys=True)
    elif data_type == "yaml":
        YAML(typ="safe").dump(configs, ostream)
        raw_result = ostream.getvalue()
        result = raw_result if raw_result else ""
    else:
        logger.error("can't serialize '%s' data (unknown type)", data_type)
        return ""

    return result


def get_compile_output(
    entry: dict, default_type: str = DEFAULT_TYPE
) -> Tuple[str, str]:
    """
    Determine the output path and type of a compile target, from the target's
    data.
    """

    # determine the output type
    output_type = entry.get("output_type", default_type)

    # write the output
    filename = entry.get("output_path", entry["name"]) + f".{output_type}"
    return os.path.join(entry["output_dir"], filename), output_type



"""
datazen - Orchestrates the "parameterized target" capability.
"""

# built-in
import logging
from typing import Dict, List, Tuple, Optional

# internal
from datazen.targets import (
    parse_targets,
    match_target,
    resolve_target_data,
    MatchData,
)


class TargetResolver:
    """
    A class for managing resolution of literal and templated target
    definitions.
    """

    def __init__(
        self, logger: logging.Logger = logging.getLogger(__name__)
    ) -> None:
        """Constuct a new target resolver."""

        self.literals: Dict[str, dict] = {}
        self.patterns: Dict[str, dict] = {}
        self.logger = logger

    def clear(self) -> None:
        """
        Re-initialize target dataset, in case a new manifest is reloaded.
        """

        self.literals = {}

    def get_target(self, group: str, name: str) -> Optional[dict]:
        """
        Attempt to get a literal target from what has been loaded so far.
        """

        # try to return an existing literal target if possible
        assert group in self.literals and group in self.patterns
        if group in self.literals and name in self.literals[group]:
            data = self.literals[group][name]
            assert data["literal"]
            return data["data"]

        # short-circuit case where we don't have any patterns for this group
        # to try
        if group not in self.patterns or not self.patterns[group]:
            return None

        # attempt to match this target to any of our patterns for this group
        matches: List[Tuple[dict, MatchData]] = []
        for pattern in self.patterns[group].values():
            result = match_target(name, pattern["pattern"], pattern["keys"])
            if result.found:
                matches.append((pattern, result.substitutions))

        # make sure we matched only one target
        if not matches or len(matches) > 1:
            log_str = (
                "couldn't match one target for '%s-%s', found "
                + "%d candidates"
            )
            self.logger.error(log_str, group, name, len(matches))
            for match in matches:
                self.logger.error("%s", match[0]["data"]["name"])
            return None

        # create a new target from the template, save it as a new literal so
        # we don't need to re-match it
        new_literal = resolve_target_data(matches[0][0]["data"], matches[0][1])
        data = {
            "literal": True,
            "data": new_literal,
            "pattern": new_literal["name"],
            "keys": [],
        }
        data["data"]["overrides"] = matches[0][1]
        self.literals[group][new_literal["name"]] = data
        return new_literal

    def register_group(self, name: str, targets: List[dict]) -> None:
        """
        From a name of a group that contains targets, initialize it by also
        providing its target datset.
        """

        self.literals[name], self.patterns[name] = parse_targets(targets)



import logging
from datazen.compile import str_compile as str_compile
from datazen.environment.integrated import Environment as Environment
from datazen.parsing import get_file_ext as get_file_ext
from typing import Any, List

LOG: Any

def cmd_compile(config_dirs: List[str], schema_dirs: List[str], variable_dirs: List[str], output_file_path: str, logger: logging.Logger = ...) -> bool: ...



"""
datazen - A child class for adding configuration-data loading capabilities to
          the environment dataset.
"""

# built-in
import logging
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.environment.variable import VariableEnvironment
from datazen.environment.schema import SchemaEnvironment
from datazen.configs import load as load_configs
from datazen.load import LoadedFiles, DEFAULT_LOADS


class ConfigEnvironment(VariableEnvironment, SchemaEnvironment):
    """
    The configuration-data loading environment mixin, requires variable
    loading to function.
    """

    def __init__(self):
        """Extend the environment with a notion of configs being valid."""

        super().__init__()
        self.configs_valid = False

    def load_configs(
        self,
        cfg_loads: LoadedFiles = DEFAULT_LOADS,
        var_loads: LoadedFiles = DEFAULT_LOADS,
        sch_loads: LoadedFiles = DEFAULT_LOADS,
        sch_types_loads: LoadedFiles = DEFAULT_LOADS,
        name: str = ROOT_NAMESPACE,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> dict:
        """
        Load configuration data, resolve any un-loaded configuration
        directories.
        """

        self.configs_valid = False

        # determine directories that need to be loaded
        data_type = DataType.CONFIG

        with self.lock:
            to_load = self.get_to_load(data_type, name)

            # load new data
            config_data = self.get_data(data_type, name)
            if to_load:
                vdata = self.load_variables(var_loads, name)
                config_data.update(load_configs(to_load, vdata, cfg_loads))
                self.update_load_state(data_type, to_load, name)

        # enforce schemas
        if not self.enforce_schemas(
            config_data, True, sch_loads, sch_types_loads, name
        ):
            logger.error("schema validation failed, returning an empty dict")
            return {}

        self.configs_valid = True
        return config_data

    def add_config_dirs(
        self,
        dir_paths: List[str],
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> int:
        """
        Add configuration-data directories, return the number of directories
        added.
        """

        return self.add_dirs(
            DataType.CONFIG, dir_paths, rel_path, name, allow_dup
        )



import jinja2
import logging
from datazen import CACHE_SUFFIX as CACHE_SUFFIX, DEFAULT_MANIFEST as DEFAULT_MANIFEST, ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.classes.file_info_cache import FileInfoCache as FileInfoCache, cmp_total_loaded as cmp_total_loaded
from datazen.environment.manifest import ManifestEnvironment as ManifestEnvironment
from datazen.paths import get_file_name as get_file_name
from typing import Any, Dict, List

LOG: Any

def manifest_cache_dir(path: str, manifest: dict) -> str: ...

class ManifestCacheEnvironment(ManifestEnvironment):
    cache: Any
    aggregate_cache: Any
    initial_cache: Any
    manifest_changed: bool
    def __init__(self) -> None: ...
    def load_manifest_with_cache(self, path: str = ..., logger: logging.Logger = ...) -> bool: ...
    def clean_cache(self, purge_data: bool = ...) -> None: ...
    def write_cache(self) -> None: ...
    def describe_cache(self) -> None: ...
    def restore_cache(self) -> None: ...
    def get_new_loaded(self, types: List[str], load_checks: Dict[str, List[str]] = ...) -> int: ...
    def cached_load_variables(self, name: str = ...) -> dict: ...
    def cached_load_schemas(self, require_all: bool = ..., name: str = ...) -> dict: ...
    def cached_enforce_schemas(self, data: dict, require_all: bool = ..., name: str = ...) -> bool: ...
    def cached_load_configs(self, name: str = ...) -> dict: ...
    def cached_load_templates(self, name: str = ...) -> Dict[str, jinja2.Template]: ...



import logging
from datazen import ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.classes.task_data_cache import TaskDataCache as TaskDataCache
from datazen.environment.base import Task as Task, TaskResult as TaskResult, dep_slug_unwrap as dep_slug_unwrap
from datazen.environment.manifest import set_output_dir as set_output_dir
from datazen.environment.manifest_cache import ManifestCacheEnvironment as ManifestCacheEnvironment
from datazen.parsing import merge as merge
from typing import Any, Callable, Dict, List, Optional, Tuple

LOG: Any
TaskFunction = Callable[..., TaskResult]

class TaskEnvironment(ManifestCacheEnvironment):
    def valid_noop(self, entry: dict, _: str, __: dict = ..., ___: List[str] = ..., logger: logging.Logger = ...) -> TaskResult: ...
    visited: Any
    is_new: Any
    default: str
    handles: Any
    data_cache: Any
    def __init__(self): ...
    def init_cache(self, cache_dir: str) -> None: ...
    def write_cache(self) -> None: ...
    def clean_cache(self, purge_data: bool = ...) -> None: ...
    @property
    def task_data(self) -> dict: ...
    def is_resolved(self, operation: str, target: str) -> bool: ...
    def is_task_new(self, operation: str, target: str) -> bool: ...
    def resolve(self, operation: str, target: str, should_cache: bool, is_new: bool) -> None: ...
    def push_dep(self, dep: str, task_stack: List[Task], curr_target: str) -> None: ...
    def get_dep_data(self, dep_list: List[str]) -> dict: ...
    def already_satisfied(self, target: str, output_path: Optional[str], load_deps: List[str], deps_changed: List[str] = ..., load_checks: Dict[str, List[str]] = ..., logger: logging.Logger = ...) -> bool: ...
    def resolve_dependencies(self, dep_list: List[str], task_stack: List[Task], target: str) -> Tuple[bool, dict, List[str]]: ...
    def get_manifest_entry(self, category: str, name: str) -> dict: ...
    def handle_task(self, key_name: str, target: str, task_stack: List[Task] = ..., should_cache: bool = ...) -> TaskResult: ...

def get_dep_list(entry: dict) -> List[str]: ...
def get_path(entry: dict, key: str = ...) -> str: ...



import logging
from datazen import CACHE_SUFFIX as CACHE_SUFFIX
from datazen.environment.base import dep_slug_unwrap as dep_slug_unwrap
from datazen.environment.command import CommandEnvironment as CommandEnvironment
from datazen.environment.compile import CompileEnvironment as CompileEnvironment
from datazen.environment.group import GroupEnvironment as GroupEnvironment
from datazen.environment.render import RenderEnvironment as RenderEnvironment
from typing import Any, List, Tuple

LOG: Any

class Environment(CompileEnvironment, RenderEnvironment, GroupEnvironment, CommandEnvironment):
    visited: Any
    default: str
    logger: Any
    def __init__(self, logger: logging.Logger = ...) -> None: ...
    def execute(self, target: str = ..., should_cache: bool = ...) -> Tuple[bool, bool]: ...
    def execute_targets(self, targets: List[str]) -> bool: ...
    def group(self, target: str) -> Tuple[bool, bool]: ...
    def compile(self, target: str) -> Tuple[bool, bool]: ...
    def render(self, target: str) -> Tuple[bool, bool]: ...
    def command(self, target: str) -> Tuple[bool, bool]: ...

def from_manifest(manifest_path: str, data_cache_name: str = ..., logger: logging.Logger = ...) -> Environment: ...



"""
datazen - A class for adding manifest-loading to environments.
"""

# built-in
import logging
import os
from io import StringIO
from typing import List, Tuple

# third-party
from cerberus import Validator

# internal
from datazen.environment.config import ConfigEnvironment
from datazen.classes.target_resolver import TargetResolver
from datazen.environment.template import TemplateEnvironment
from datazen.parsing import load as load_raw
from datazen.parsing import load_stream, merge_dicts
from datazen.paths import get_package_data, get_package_dir, resolve_dir
from datazen.schemas import load_types, inject_custom_schemas
from datazen import DEFAULT_DIR, ROOT_NAMESPACE

LOG = logging.getLogger(__name__)


def get_output_dir(
    data: dict, rel_path: str, default: str = DEFAULT_DIR
) -> str:
    """
    Get the resolved output directory based on a dictionary containing
    target data.
    """

    # turn the output directory into a valid path
    out_dir = default
    if "output_dir" in data:
        out_dir = data["output_dir"]
    return resolve_dir(out_dir, rel_path)


def set_output_dir(
    data: dict, rel_path: str, default: str = DEFAULT_DIR
) -> None:
    """Set the 'output_dir' key correctly on a dictionary."""

    out_dir = get_output_dir(data, rel_path, default)
    os.makedirs(out_dir, exist_ok=True)
    data["output_dir"] = out_dir


class ManifestEnvironment(ConfigEnvironment, TemplateEnvironment):
    """
    A wrapper for the manifest-loading implementations of an environment.
    """

    def __init__(self):
        """Add a manifest dictionary to the environment."""

        super().__init__()
        self.manifest = {}
        self.target_resolver = TargetResolver()

    def load_dirs(
        self,
        data: dict,
        rel_path: str,
        namespace: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> None:
        """
        Looks for keys matching types of directories that can be loaded
        into an environment and tries to load them.
        """

        key_handles = {
            "configs": self.add_config_dirs,
            "schemas": self.add_schema_dirs,
            "schema_types": self.add_schema_type_dirs,
            "templates": self.add_template_dirs,
            "variables": self.add_variable_dirs,
        }
        for key, handler in key_handles.items():
            if key in data:
                handler(data[key], rel_path, namespace, allow_dup)
            # if a directory list isn't provided, and the directory of the
            # same name of the key is present in the manifest directory,
            # load it
            elif os.path.isdir(os.path.join(rel_path, key)):
                handler([key], rel_path, namespace, allow_dup)

    def default_target(self) -> str:
        """
        Retrieve a configured default target for a manifest, otherwise an
        empty String.
        """

        result = ""
        if self.get_valid() and "default_target" in self.manifest["data"]:
            result = self.manifest["data"]["default_target"]
        return result

    def load_manifest_reent(
        self,
        path: str,
        manifest_dir: str,
        params: dict,
        files: List[str],
        logger: logging.Logger = LOG,
    ) -> Tuple[dict, bool]:
        """
        Load a manifest recursively by resolving includes and merging the
        results.
        """

        if not os.path.isabs(path):
            path = os.path.join(manifest_dir, path)
        files.append(os.path.abspath(path))

        # load raw data
        curr_manifest, loaded = load_raw(path, params, {})

        # update params, load again so we can use self-referential params
        if "params" in curr_manifest:
            params = merge_dicts([params, curr_manifest["params"]])
            curr_manifest, loaded = load_raw(path, params, {})

        if not loaded:
            return curr_manifest, False

        # load the data directories before resolving includes
        rel_path = os.path.dirname(path)
        self.load_dirs(curr_manifest, rel_path, allow_dup=True)

        # resolve includes
        all_manifests = [curr_manifest]
        if "includes" in curr_manifest:
            for include in curr_manifest["includes"]:
                result = self.load_manifest_reent(
                    include, rel_path, params, files, logger
                )
                if not result[1]:
                    logger.info("include '%s' failed to load", path)
                    return curr_manifest, False
                all_manifests.append(result[0])

        # merge include data
        curr_manifest = merge_dicts(all_manifests)

        return curr_manifest, True

    def load_manifest(
        self, path: str = "manifest.yaml", logger: logging.Logger = LOG
    ) -> bool:
        """Attempt to load manifest data from a file."""

        # don't allow double-loading manifests
        if self.manifest:
            logger.error(
                "manifest '%s' already loaded for this environment",
                self.manifest["path"],
            )
            return False

        path = os.path.abspath(path)
        manifest_dir = os.path.dirname(path)
        self.manifest["path"] = path
        self.manifest["dir"] = manifest_dir
        files: List[str] = []
        self.manifest["data"], loaded = self.load_manifest_reent(
            path, manifest_dir, {}, files, logger
        )
        self.manifest["files"] = files

        # set up target resolver
        self.target_resolver.clear()
        candidates = ["compiles", "commands", "renders", "groups"]
        for candidate in candidates:
            if candidate in self.manifest["data"]:
                cand_data = self.manifest["data"][candidate]
                self.target_resolver.register_group(candidate, cand_data)

        # make sure we loaded a manifest
        if not loaded:
            self.set_valid(False)
            return self.get_valid()

        # enforce the manifest schema
        if not validate_manifest(self.manifest["data"], logger):
            self.set_valid(False)
            return self.get_valid()

        # add directories parsed from the schema, paths in the manifest are
        # relative to the directory the manifest is located
        rel_path = self.manifest["dir"]
        set_output_dir(self.manifest["data"], rel_path)

        return self.get_valid()


def get_manifest_schema(
    require_all: bool = True,
    logger: logging.Logger = LOG,
) -> Validator:
    """Load the schema for manifest from the package."""

    rel_path = os.path.join("schemas", "manifest.yaml")
    schema_str = get_package_data(rel_path)
    return Validator(
        load_stream(StringIO(schema_str), rel_path, logger).data,
        require_all=require_all,
    )


def validate_manifest(manifest: dict, logger: logging.Logger = LOG) -> bool:
    """Validate manifest data against the package schema."""

    custom_schemas = load_types([get_package_dir("schema_types")])

    with inject_custom_schemas(custom_schemas):
        schema = get_manifest_schema(False, logger)
        result = schema.validate(manifest)

    if not result:
        logger.error("invalid manifest: %s", schema.errors)

    return result



"""
datazen - An environment extension that exposes compilation capabilities.
"""

# built-in
import logging
from typing import List

# internal
from datazen.compile import str_compile, get_compile_output
from datazen.environment.base import TaskResult
from datazen.environment.task import TaskEnvironment
from datazen.parsing import merge_dicts
from datazen.paths import advance_dict_by_path
from datazen.targets import resolve_dep_data

LOG = logging.getLogger(__name__)


class CompileEnvironment(TaskEnvironment):
    """Leverages a cache-equipped environment to perform compilations."""

    def __init__(self):
        """
        Add compiled data to a dictionary whenever a compilation is performed.
        """

        super().__init__()
        self.handles["compiles"] = self.valid_compile

    def valid_compile(
        self,
        entry: dict,
        namespace: str,
        dep_data: dict = None,
        deps_changed: List[str] = None,
        logger: logging.Logger = LOG,
    ) -> TaskResult:
        """Perform the compilation specified by the entry."""

        path, output_type = get_compile_output(entry)

        # load configs early to update cache
        data = self.cached_load_configs(namespace)

        # update this dict with the dependency data
        if dep_data is not None:
            if entry.get("merge_deps", False):
                data = merge_dicts([{}, data, dep_data], logger=logger)

            # this isn't a good default behavior in practice, but older code
            # (and tests) rely on it
            else:
                data.update(dep_data)

        # advance the dict if it was requested
        if "index_path" in entry:
            data = advance_dict_by_path(entry["index_path"].split("."), data)

        # apply overrides if present
        data = resolve_dep_data(entry, data)

        # set task-data early, in case we don't need to re-compile
        self.task_data["compiles"][entry["name"]] = data

        # make sure this compilation needs to be performed
        if self.already_satisfied(
            entry["name"],
            path,
            ["configs", "variables", "schemas"],
            deps_changed,
        ):
            logger.debug("compile '%s' satisfied, skipping", entry["name"])
            return TaskResult(True, False)

        mode = "a" if "append" in entry and entry["append"] else "w"
        with open(path, mode, encoding="utf-8") as out_file:
            out_file.write(str_compile(data, output_type))
            logger.info("compiled '%s' data to '%s'", output_type, path)

        return TaskResult(True, True)



# =====================================
# generator=datazen
# version=1.7.8
# hash=4c8620a30a13b93abde0ea0c3c0e156a
# =====================================

"""
vmklib - Useful defaults and other package metadata.
"""

PKG_NAME = "vmklib"
VERSION = "0.5.0"
DESCRIPTION = "Simplify project workflows by standardizing use of GNU Make."



"""
datazen - Top-level APIs for loading and interacting with configuration data.
"""

# built-in
from typing import List

# internal
from datazen.load import load_dir, LoadedFiles, DEFAULT_LOADS


def load(
    directories: List[str],
    variable_data: dict = None,
    loads: LoadedFiles = DEFAULT_LOADS,
) -> dict:
    """Load configuration data from a list of directories."""

    result: dict = {}
    for directory in directories:
        load_dir(directory, result, variable_data, loads)
    return result



"""
datazen - A class for storing data from completed operations to disk.
"""

# built-in
from collections import defaultdict
import os
import shutil

# internal
from datazen.compile import write_dir
from datazen.load import load_dir_only


class TaskDataCache:
    """
    Provides storage for data produced by task targets to facilitate better
    (and more correct) short-circuiting.
    """

    def __init__(self, cache_dir: str):
        """Construct an empty cache or optionally load from a directory."""

        self.data: dict = defaultdict(lambda: defaultdict(dict))
        self.cache_dir = cache_dir
        self.load(self.cache_dir)

    def load(self, load_dir: str) -> None:
        """Read new data from the cache directory and update state."""

        os.makedirs(load_dir, exist_ok=True)
        self.data.update(load_dir_only(load_dir, are_templates=False))

    def save(self, out_type: str = "json") -> None:
        """Write cache data to disk."""

        write_dir(self.cache_dir, self.data, out_type)

    def clean(self, purge_data: bool = True) -> None:
        """Clean this cache's data on disk."""

        shutil.rmtree(self.cache_dir)
        if purge_data:
            self.data = defaultdict(lambda: defaultdict(dict))



import jinja2
from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles
from datazen.parsing import set_file_hash as set_file_hash
from datazen.paths import get_file_ext as get_file_ext, get_file_name as get_file_name
from typing import Dict, List

def update_cache_primitives(dir_path: str, loads: LoadedFiles) -> None: ...
def load(template_dirs: List[str], loads: LoadedFiles = ...) -> Dict[str, jinja2.Template]: ...



"""
datazen - Top-level APIs for loading and interacting with schema definitions.
"""

# built-in
from contextlib import contextmanager
from typing import List, Dict, Iterator
import logging

# third-party
from cerberus import Validator, rules_set_registry

# internal
from datazen.load import load_dir, LoadedFiles, DEFAULT_LOADS
from datazen.classes.valid_dict import ValidDict

LOG = logging.getLogger(__name__)


def load(
    directories: List[str],
    require_all: bool = True,
    loads: LoadedFiles = DEFAULT_LOADS,
) -> dict:
    """Load schemas from a list of directories."""

    result: dict = {}

    # load raw data
    for directory in directories:
        load_dir(directory, result, None, loads)

    # interpret all top-level keys as schemas
    schemas = {}
    for item in result.items():
        schemas[item[0]] = Validator(item[1], require_all=require_all)
    return schemas


def add_global_schemas(
    schema_data: Dict[str, dict], logger: logging.Logger = LOG
) -> None:
    """Add schema-type registrations, globally."""

    for key, schema in schema_data.items():
        logger.debug("adding '%s' schema type", key)
        rules_set_registry.add(key, schema)


def remove_global_schemas(
    schema_data: Dict[str, dict], logger: logging.Logger = LOG
) -> None:
    """Remove schema-type registrations by key name."""

    if schema_data:
        logger.debug(
            "removing schema types '%s'", ", '".join(schema_data.keys())
        )
        rules_set_registry.remove(*schema_data.keys())


def load_types(
    directories: List[str],
    loads: LoadedFiles = DEFAULT_LOADS,
) -> Dict[str, dict]:
    """Load schema types and optionally register them."""

    schema_data: Dict[str, dict] = {}

    # load raw data
    for directory in directories:
        load_dir(directory, schema_data, None, loads)
    return schema_data


def validate(
    schema_data: Dict[str, Validator],
    data: dict,
    logger: logging.Logger = LOG,
) -> bool:
    """
    For every top-level key in the schema data, attempt to validate the
    provided data.
    """

    for item in schema_data.items():
        key = item[0]
        if key in data:
            if not ValidDict(key, data[key], item[1], logger).valid:
                return False

    # warn if anything wasn't validated
    for item in data.keys():
        if item not in schema_data:
            logger.warning("no schema for '%s', not validating", item)

    return True


@contextmanager
def inject_custom_schemas(
    schema_data: Dict[str, dict],
    should_inject: bool = True,
    logger: logging.Logger = LOG,
) -> Iterator[None]:
    """
    Allow the user to more easily control adding and removing global schema
    definitions.
    """

    try:
        if should_inject:
            add_global_schemas(schema_data, logger)
        yield
    finally:
        if should_inject:
            remove_global_schemas(schema_data, logger)



import re
from datazen.parsing import merge as merge
from datazen.paths import advance_dict_by_path as advance_dict_by_path, format_resolve_delims as format_resolve_delims, unflatten_dict as unflatten_dict
from typing import Dict, List, NamedTuple, Tuple

KW_OPEN: str
KW_CLOSE: str
KW_PATTERN: str

def target_is_literal(name: str) -> bool: ...

class ParseResult(NamedTuple):
    pattern: re.Pattern
    keys: List[str]

def parse_target(name: str) -> ParseResult: ...
def parse_targets(targets: List[dict]) -> Tuple[Dict[str, dict], Dict[str, dict]]: ...
MatchData = Dict[str, str]

class TargetMatch(NamedTuple):
    found: bool
    substitutions: MatchData

def match_target(name: str, pattern: re.Pattern, keys: List[str]) -> TargetMatch: ...
def resolve_target_list(target_list: list, match_data: MatchData) -> list: ...
def resolve_dep_data(entry: dict, data: dict) -> dict: ...
def resolve_target_data(target_data: dict, match_data: MatchData) -> dict: ...



import logging
from datazen.environment.base import TaskResult as TaskResult
from datazen.environment.task import TaskEnvironment as TaskEnvironment
from typing import List

class GroupEnvironment(TaskEnvironment):
    def __init__(self) -> None: ...
    def valid_group(self, entry: dict, _: str, dep_data: dict = ..., deps_changed: List[str] = ..., logger: logging.Logger = ...) -> TaskResult: ...



"""
datazen - APIs for loading raw data from files.
"""

# built-in
import hashlib
import io
import json
import logging
from typing import TextIO, List, NamedTuple
import time

# third-party
import jinja2
from ruamel.yaml import YAML, scanner, parser

# internal
from datazen.paths import get_file_ext

LOG = logging.getLogger(__name__)


class LoadResult(NamedTuple):
    """
    An encapsulation of the result of loading raw data, the data collected and
    whether or not it succeeded.
    """

    data: dict
    success: bool


def get_json_data(
    data_file: TextIO,
    logger: logging.Logger = LOG,
) -> LoadResult:
    """Load JSON data from a text stream."""

    data = {}
    loaded = True
    try:
        data = json.load(data_file)
        if not data:
            data = {}
    except json.decoder.JSONDecodeError as exc:
        loaded = False
        logger.error("json-load error: %s", exc)
    return LoadResult(data, loaded)


def get_yaml_data(
    data_file: TextIO,
    logger: logging.Logger = LOG,
) -> LoadResult:
    """Load YAML data from a text stream."""

    data = {}
    loaded = True
    try:
        data = YAML(typ="safe").load(data_file)
        if not data:
            data = {}
    except (scanner.ScannerError, parser.ParserError) as exc:
        loaded = False
        logger.error("yaml-load error: %s", exc)
    return LoadResult(data, loaded)


def load_stream(
    data_stream: TextIO,
    data_path: str,
    logger: logging.Logger = LOG,
) -> LoadResult:
    """
    Load arbitrary data from a text stream, update an existing dictionary.
    """

    # update the dictionary
    ext = get_file_ext(data_path)
    result = LoadResult({}, False)
    if ext == "json":
        result = get_json_data(data_stream, logger)
    elif ext == "yaml":
        result = get_yaml_data(data_stream, logger)
    else:
        logger.error(
            "can't load data from '%s' (unknown extension '%s')",
            data_path,
            ext,
        )

    if not result.success:
        logger.error("failed to load '%s'", data_path)

    return result


def dedup_dict_lists(data: dict) -> dict:
    """
    Finds list elements in a dictionary and removes duplicate entries, mutates
    the original list.
    """

    for key in data:
        if isinstance(data[key], dict):
            data[key] = dedup_dict_lists(data[key])
        elif isinstance(data[key], list):
            new_list = []
            for item in data[key]:
                if item not in new_list:
                    new_list.append(item)
            data[key] = new_list

    return data


def merge_dicts(
    dicts: List[dict],
    expect_overwrite: bool = False,
    logger: logging.Logger = LOG,
) -> dict:
    """
    Merge a list of dictionary data into a single set (mutates the first
    element).
    """

    result = dicts[0]
    for right_dict in dicts[1:]:
        result = merge(
            result,
            right_dict,
            expect_overwrite=expect_overwrite,
            logger=logger,
        )
    return result


def merge(
    dict_a: dict,
    dict_b: dict,
    path: List[str] = None,
    expect_overwrite: bool = False,
    logger: logging.Logger = LOG,
) -> dict:
    """
    Combine two dictionaries recursively, prefers dict_a in a conflict. For
    values of the same key that are lists, the lists are combined. Otherwise
    the resulting dictionary is cleanly merged.
    """

    if path is None:
        path = []

    for key in dict_b:
        if key in dict_a:
            # first try to coerce b's type into a's
            if not isinstance(dict_b[key], type(dict_a[key])):
                try:
                    dict_b[key] = type(dict_a[key])(dict_b[key])
                except ValueError:
                    pass

            # same leaf value
            if dict_a[key] == dict_b[key]:
                pass
            elif isinstance(dict_a[key], dict) and isinstance(
                dict_b[key], dict
            ):
                merge(
                    dict_a[key],
                    dict_b[key],
                    path + [str(key)],
                    expect_overwrite,
                )
            elif isinstance(dict_a[key], list) and isinstance(
                dict_b[key], list
            ):
                dict_a[key].extend(dict_b[key])
            elif not isinstance(dict_b[key], type(dict_a[key])):
                logger.error(
                    "Type mismatch at '%s'", ".".join(path + [str(key)])
                )
                logger.error("left:  %s (%s)", type(dict_a[key]), dict_a[key])
                logger.error("right: %s (%s)", type(dict_b[key]), dict_b[key])
            elif not expect_overwrite:
                logger.error("Conflict at '%s'", ".".join(path + [str(key)]))
                logger.error("left:  %s", dict_a[key])
                logger.error("right: %s", dict_b[key])
            else:
                dict_a[key] = dict_b[key]
        else:
            dict_a[key] = dict_b[key]

    return dict_a


def load(
    data_path: str,
    variables: dict,
    dict_to_update: dict,
    expect_overwrite: bool = False,
    is_template: bool = True,
    logger: logging.Logger = LOG,
) -> LoadResult:
    """
    Load raw file data and meld it into an existing dictionary. Update
    the result as if it's a template using the provided variables.
    """

    result = LoadResult({}, False)

    # read the raw file and interpret it as a template, resolve 'variables'
    try:
        with open(data_path, encoding="utf-8") as config_file:
            if is_template:
                template = jinja2.Template(config_file.read())
                str_output = template.render(variables)
            else:
                str_output = config_file.read()
    except FileNotFoundError:
        logger.error("can't find '%s' to load file data", data_path)
        return result
    except jinja2.exceptions.TemplateError as exc:
        logger.error(
            "couldn't render '%s': %s (variables: %s)",
            data_path,
            exc,
            variables,
        )
        return result

    new_data, loaded = load_stream(io.StringIO(str_output), data_path, logger)
    return LoadResult(
        merge(dict_to_update, new_data, expect_overwrite=expect_overwrite),
        loaded,
    )


def get_hash(data: str, encoding: str = "utf-8") -> str:
    """Get the MD5 of String data."""

    return hashlib.md5(bytes(data, encoding)).hexdigest()


def get_file_hash(path: str) -> str:
    """Get the MD5 of a file by path."""

    with open(path, encoding="utf-8") as data:
        contents = data.read()
    return get_hash(contents)


def set_file_hash(hashes: dict, path: str, set_new: bool = True) -> bool:
    """Evaluate a hash dictionary and update it on a miss."""

    str_hash = get_file_hash(path)
    result = True
    if path in hashes and str_hash == hashes[path]["hash"]:
        result = False
    elif set_new:
        if path not in hashes:
            hashes[path] = {}
        hashes[path]["hash"] = str_hash
        hashes[path]["time"] = time.time()

    return result



import git
import logging
from datazen.compile import write_dir as write_dir
from datazen.load import load_dir_only as load_dir_only
from datazen.parsing import merge as merge
from datazen.paths import EXCLUDES as EXCLUDES
from typing import Any, Iterator

class DataRepository:
    root: Any
    out_type: Any
    repo: Any
    data: Any
    lock: Any
    logger: Any
    def __init__(self, root_dir: str, out_type: str = ..., logger: logging.Logger = ...) -> None: ...
    def meld(self, data: dict, root_rel: str = ..., expect_overwrite: bool = ...) -> Iterator[git.Repo]: ...
    def loaded(self, root_rel: str = ..., write_back: bool = ...) -> Iterator[dict]: ...



"""
datazen - An interface for managing on-disk data by loading it and making
          discrete changes.
"""

# built-in
from contextlib import contextmanager
import logging
import os
import threading
from typing import Iterator

# third-party
import git

# internal
from datazen.compile import write_dir
from datazen.load import load_dir_only
from datazen.parsing import merge
from datazen.paths import EXCLUDES


class DataRepository:
    """
    A class for interacting with file-backed databases that are built with
    serialization formats supported by this package.
    """

    def __init__(
        self,
        root_dir: str,
        out_type: str = "yaml",
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        """Construct a new data repository."""

        self.root: str = os.path.realpath(root_dir)
        self.out_type = out_type
        self.repo = git.Repo(self.root)
        # we should store some useful state - maybe what branch / head we're on
        # we may even want to split out the git stuff into its own class
        self.data: dict = {}
        self.lock = threading.RLock()
        self.logger = logger

    @contextmanager
    def meld(
        self, data: dict, root_rel: str = ".", expect_overwrite: bool = False
    ) -> Iterator[git.Repo]:
        """
        Update the data at some root-relative path, write it to disk and then
        provided the repository object still within a locked context.
        """

        with self.lock:
            # update data with contents provided
            with self.loaded(root_rel) as curr_data:
                self.data = merge(
                    curr_data, data, expect_overwrite=expect_overwrite
                )

            # yield the repository after we've written disk but before
            # releasing the lock
            yield self.repo

    @contextmanager
    def loaded(
        self, root_rel: str = ".", write_back: bool = True
    ) -> Iterator[dict]:
        """
        Load a data directory, yield it to the caller and write it back when
        complete inside a locked context for this repository.
        """

        to_load = os.path.realpath(os.path.join(self.root, root_rel))
        assert os.path.isdir(to_load)
        with self.lock:
            self.data = load_dir_only(to_load)
            self.logger.info("loading '%s'", to_load)
            yield self.data
            if write_back:
                # remove items that contain data but not others
                for item in os.listdir(to_load):
                    if item not in EXCLUDES:
                        os.remove(os.path.join(to_load, item))
                write_dir(to_load, self.data, self.out_type)
                self.logger.info("writing '%s'", to_load)



import jinja2
import logging
from datazen.environment.integrated import Environment as Environment
from datazen.parsing import load_stream as load_stream
from typing import Any, List

LOG: Any

def str_render(template: jinja2.Template, config_data_path: str, logger: logging.Logger = ...) -> str: ...
def cmd_render(template_dirs: List[str], template_name: str, config_data_path: str, output_file_path: str, logger: logging.Logger = ...) -> bool: ...



"""
datazen - A child class for adding schema-data loading capabilities to the
          environment dataset.
"""

# built-in
from typing import List

# internal
from datazen import ROOT_NAMESPACE
from datazen.enums import DataType
from datazen.environment.base import BaseEnvironment
from datazen.load import LoadedFiles, DEFAULT_LOADS
from datazen.schemas import load as load_schemas
from datazen.schemas import validate, load_types, inject_custom_schemas


class SchemaEnvironment(BaseEnvironment):
    """
    The schema-data loading environment mixin, only requires the base
    environment capability to function.
    """

    def load_schema_types(
        self,
        sch_loads: LoadedFiles = DEFAULT_LOADS,
        name: str = ROOT_NAMESPACE,
    ) -> dict:
        """Load custom schema types and resolve any un-loaded directories."""

        # determine directories that need to be loaded
        data_type = DataType.SCHEMA_TYPES

        with self.lock:
            to_load = self.get_to_load(data_type, name)

            # load schema data for this namespace
            schema_type_data = self.get_data(data_type, name)
            if to_load:
                schema_type_data.update(load_types(to_load, sch_loads))

        return schema_type_data

    def load_schemas(
        self,
        require_all: bool = True,
        sch_loads: LoadedFiles = DEFAULT_LOADS,
        sch_types_loads: LoadedFiles = DEFAULT_LOADS,
        name: str = ROOT_NAMESPACE,
        modify_registry: bool = True,
    ) -> dict:
        """Load schema data, resolve any un-loaded schema directories."""

        # determine directories that need to be loaded
        data_type = DataType.SCHEMA

        sch_types = self.load_schema_types(sch_types_loads, name)

        with self.lock:
            to_load = self.get_to_load(data_type, name)

            # load new data
            schema_data = self.get_data(data_type, name)
            if to_load:
                with inject_custom_schemas(sch_types, modify_registry):
                    schema_data.update(
                        load_schemas(to_load, require_all, sch_loads)
                    )

        return schema_data

    def enforce_schemas(
        self,
        data: dict,
        require_all: bool = True,
        sch_loads: LoadedFiles = DEFAULT_LOADS,
        sch_types_loads: LoadedFiles = DEFAULT_LOADS,
        name: str = ROOT_NAMESPACE,
    ) -> bool:
        """
        Perform schema-validation on provided data and return the boolean
        result. Adds (and removes) namespaced types if applicable.
        """

        with self.lock:
            sch_types = self.load_schema_types(sch_types_loads, name)
            with inject_custom_schemas(sch_types):
                result = validate(
                    self.load_schemas(
                        require_all, sch_loads, sch_types_loads, name, False
                    ),
                    data,
                )

        return result

    def add_schema_type_dirs(
        self,
        dir_paths: List[str],
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> int:
        """
        Add directories containing schema-type data (to be registered at
        runtime).
        """

        return self.add_dirs(
            DataType.SCHEMA_TYPES, dir_paths, rel_path, name, allow_dup
        )

    def add_schema_dirs(
        self,
        dir_paths: List[str],
        rel_path: str = ".",
        name: str = ROOT_NAMESPACE,
        allow_dup: bool = False,
    ) -> int:
        """
        Add schema-data directories, return the number of directories added.
        """

        return self.add_dirs(
            DataType.SCHEMA, dir_paths, rel_path, name, allow_dup
        )



"""
datazen - A class for exposing the capability to run concrete tasks against
          the current environment.
"""

# built-in
from collections import defaultdict
import logging
import os
from typing import Callable, List, Tuple, Dict, Optional

# internal
from datazen import ROOT_NAMESPACE
from datazen.environment.base import dep_slug_unwrap, Task, TaskResult
from datazen.environment.manifest import set_output_dir
from datazen.environment.manifest_cache import ManifestCacheEnvironment
from datazen.classes.task_data_cache import TaskDataCache
from datazen.parsing import merge

LOG = logging.getLogger(__name__)

TaskFunction = Callable[..., TaskResult]


class TaskEnvironment(ManifestCacheEnvironment):
    """
    Adds capability for running tasks (including tasks with dependencies)
    and tracking whether or not updates are required.
    """

    def valid_noop(
        self,
        entry: dict,
        _: str,
        __: dict = None,
        ___: List[str] = None,
        logger: logging.Logger = LOG,
    ) -> TaskResult:
        """Default handle for a potentially-unregistered operation."""

        logger.error(
            "'%s' no handler found (default: '%s')",
            entry["name"],
            self.default,
        )
        return TaskResult(False, False)

    def __init__(self):
        """Add a notion of 'visited' targets to the environment data."""

        super().__init__()
        self.visited = defaultdict(bool)
        self.is_new = defaultdict(bool)
        self.default = "noop"
        self.handles: Dict[str, TaskFunction] = defaultdict(
            lambda: self.valid_noop
        )
        self.data_cache: Optional[TaskDataCache] = None

    def init_cache(self, cache_dir: str) -> None:
        """Initialize the task-data cache."""

        if self.data_cache is None:
            self.data_cache = TaskDataCache(cache_dir)

    def write_cache(self) -> None:
        """Commit cached data to the file-system."""

        super().write_cache()
        if self.data_cache is not None:
            self.data_cache.save()

    def clean_cache(self, purge_data: bool = True) -> None:
        """Remove cached data from the file-system."""

        super().clean_cache(True)
        if self.data_cache is not None:
            self.data_cache.clean(purge_data)

    @property
    def task_data(self) -> dict:
        """Proxy task data through the cache."""

        assert self.data_cache is not None
        return self.data_cache.data

    def is_resolved(self, operation: str, target: str) -> bool:
        """
        Determine whether a target for an operation has already been resolved.
        """

        with self.lock:
            return self.visited[Task(operation, target).slug]

    def is_task_new(self, operation: str, target: str) -> bool:
        """Determine if a target has been newly executed this iteration."""

        with self.lock:
            return self.is_new[Task(operation, target).slug]

    def resolve(
        self, operation: str, target: str, should_cache: bool, is_new: bool
    ) -> None:
        """Set a target for an operation as resolved."""

        task = Task(operation, target)
        slug = task.slug
        with self.lock:
            self.visited[slug] = True
            self.is_new[slug] = is_new
            if should_cache:
                self.write_cache()

    def push_dep(
        self, dep: str, task_stack: List[Task], curr_target: str
    ) -> None:
        """Push a dependency onto a stack if they need to be resolved."""

        task = dep_slug_unwrap(dep, self.default)
        is_new = self.is_task_new(task.variant, task.name)
        is_resolved = self.is_resolved(task.variant, task.name)
        if dep != curr_target and (not is_resolved or is_new):
            task_stack.append(task)

    def get_dep_data(self, dep_list: List[str]) -> dict:
        """
        From a list of dependencies, create a dictionary with any task data
        they've saved.
        """

        dep_data: dict = {}

        # flatten all of the tasks' data into a single dict
        for dep in dep_list:
            task = dep_slug_unwrap(dep, self.default)

            with self.lock:
                curr_data = self.task_data[task.variant][task.name]

            if isinstance(curr_data, dict):
                dep_data = merge(dep_data, curr_data)

        return dep_data

    def already_satisfied(
        self,
        target: str,
        output_path: Optional[str],
        load_deps: List[str],
        deps_changed: List[str] = None,
        load_checks: Dict[str, List[str]] = None,
        logger: logging.Logger = LOG,
    ) -> bool:
        """
        Check if a target is already satisfied, if not debug-log some
        information about why not.
        """

        is_file = True if output_path is None else os.path.isfile(output_path)
        with self.lock:
            newly_loaded = self.get_new_loaded(load_deps, load_checks)
            result = (
                not self.manifest_changed
                and is_file
                and not deps_changed
                and newly_loaded == 0
            )

        # log less-obvious dependency-resolution hits
        if is_file:
            if deps_changed:
                logger.debug("%s: %s updates detected", target, deps_changed)
            if newly_loaded != 0:
                logger.debug(
                    "%s: %d file updates detected", target, newly_loaded
                )

        return result

    def resolve_dependencies(
        self,
        dep_list: List[str],
        task_stack: List[Task],
        target: str,
    ) -> Tuple[bool, dict, List[str]]:
        """
        Execute the entire chain of dependencies for a task, return the
        aggregate result as a boolean as well as the dependency data and which
        dependencies changed.
        """

        deps_changed = []
        for dep in dep_list:
            self.push_dep(dep, task_stack, target)

            # resolve dependencies
            while task_stack:
                task = task_stack.pop()
                result = self.handle_task(
                    task.variant, task.name, task_stack, False
                )

                # if a dependency failed, propagate it up
                if not result.success:
                    return False, {}, []
                # otherwise if a dependency was updated
                # (performed), capture it in a list
                if result.fresh or self.is_task_new(task.variant, task.name):
                    deps_changed.append(task.slug)

        # provide dependency data as "flattened"
        return True, self.get_dep_data(dep_list), deps_changed

    def get_manifest_entry(self, category: str, name: str) -> dict:
        """Get an entry from the manifest."""

        result: dict = defaultdict(lambda: None)

        with self.lock:
            candidate = self.target_resolver.get_target(category, name)
        if candidate is not None:
            result = candidate

        return result

    def handle_task(
        self,
        key_name: str,
        target: str,
        task_stack: List[Task] = None,
        should_cache: bool = True,
    ) -> TaskResult:
        """
        Handle the setup for manifest tasks, such as loading additional data
        directories and setting the correct output directory.
        """

        task = Task(key_name, target)
        slug = task.slug
        logger = logging.getLogger(slug)

        # fall back on default behavior if the manifest doesn't even have
        # an entry for this key
        if key_name not in self.manifest["data"]:
            return self.handles[key_name](
                {"name": ""}, ROOT_NAMESPACE, logger=logger
            )

        if task_stack is None:
            task_stack = []

        data = self.get_manifest_entry(key_name, target)

        if data["name"] is None:
            logger.error("no '%s' found", slug)
            return TaskResult(False, False)

        if self.is_resolved(key_name, target):
            return TaskResult(True, False)

        logger.debug("executing '%s'", slug)

        # push dependencies
        dep_result = self.resolve_dependencies(
            get_dep_list(data), task_stack, target
        )
        if not dep_result[0]:
            return TaskResult(False, False)

        # resolve the output directory
        set_output_dir(
            data, self.manifest["dir"], self.manifest["data"]["output_dir"]
        )

        # load additional data directories if specified
        with self.lock:
            namespace = self.get_namespace(key_name, target, data)
            self.load_dirs(data, self.manifest["dir"], namespace, True)

        # make sure this namespace is valid
        if not self.get_valid(namespace):
            logger.info("namespace '%s' is invalid!", namespace)
            return TaskResult(False, False)

        # write-through to the cache when we complete an operation,
        # if it succeeded
        result = self.handles[key_name](
            data, namespace, dep_result[1], dep_result[2], logger=logger
        )
        if result.success:
            self.resolve(key_name, target, should_cache, result.fresh)
        return result


def get_dep_list(entry: dict) -> List[str]:
    """From task data, build a list of task dependencies."""

    result = []
    dep_keys = ["dependencies", "children"]
    for key in dep_keys:
        if key in entry:
            result += entry[key]
    return list(set(result))


def get_path(entry: dict, key: str = "name") -> str:
    """Get the full path to a render output from the manifest entry."""

    path = entry[key]
    if "output_path" in entry:
        path = entry["output_path"]
    if not os.path.isabs(path):
        path = os.path.join(entry["output_dir"], path)
    return path



import argparse
from datazen import DEFAULT_MANIFEST as DEFAULT_MANIFEST
from datazen.environment.integrated import from_manifest as from_manifest

def entry(args: argparse.Namespace) -> int: ...
def add_app_args(parser: argparse.ArgumentParser) -> None: ...



from datazen import ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.enums import DataType as DataType
from datazen.environment import EnvironmentNamespace as EnvironmentNamespace, clone as clone
from typing import Any, List, NamedTuple

LOG: Any
SLUG_DELIM: str

class Task(NamedTuple):
    variant: str
    name: str
    @property
    def slug(self) -> str: ...

class TaskResult(NamedTuple):
    success: bool
    fresh: bool

def dep_slug_unwrap(slug: str, default_op: str) -> Task: ...

class BaseEnvironment:
    namespaces: Any
    lock: Any
    def __init__(self, default_ns: str = ...) -> None: ...
    def add_namespace(self, name: str, clone_root: bool = ...) -> None: ...
    def get_valid(self, name: str = ...) -> bool: ...
    def set_valid(self, value: bool, name: str = ...) -> None: ...
    def get_to_load(self, dir_type: DataType, name: str = ...) -> List[str]: ...
    def get_data(self, dir_type: DataType, name: str = ...) -> dict: ...
    def unload_all(self, name: str = ...) -> None: ...
    def update_load_state(self, dir_type: DataType, to_load: List[str], name: str = ...) -> int: ...
    def add_dir(self, dir_type: DataType, dir_path: str, rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> bool: ...
    def add_dirs(self, dir_type: DataType, dir_paths: List[str], rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> int: ...
    def get_namespace(self, key_name: str, target: str, target_data: dict) -> str: ...



"""
datazen - An interface for parsing and matching targets.
"""

# built-in
from collections import defaultdict
from copy import deepcopy
import re
from typing import Dict, Tuple, List, NamedTuple

# internal
from datazen.paths import (
    advance_dict_by_path,
    unflatten_dict,
    format_resolve_delims,
)
from datazen.parsing import merge

KW_OPEN = "{"
KW_CLOSE = "}"
KW_PATTERN = "[a-zA-Z0-9-_.]+"


def target_is_literal(name: str) -> bool:
    """
    Determine if a named target has keywords or not (is otherwise literal).
    """

    return name.count(KW_OPEN) == name.count(KW_CLOSE) == 0


class ParseResult(NamedTuple):
    """
    An encapsulation for a regular expression and the in-order keywords that
    can be mapped to 'group' indices.
    """

    pattern: re.Pattern
    keys: List[str]


def parse_target(name: str) -> ParseResult:
    """
    From a target name, provide a compiled pattern and an in-order list of
    the names of the keyword arguments that will appear (group order).
    """

    open_len = name.count(KW_OPEN)
    assert open_len == name.count(KW_CLOSE)

    pattern = "^"
    keys = []
    for _ in range(open_len):
        start = name.index(KW_OPEN) + 1
        end = name.index(KW_CLOSE)
        pattern += name[: start - 1]
        pattern += f"({KW_PATTERN})"
        keys.append(name[start:end])
        name = name[end + 1 :]
    pattern += name + "$"

    assert len(keys) == open_len
    return ParseResult(re.compile(pattern), keys)


def parse_targets(
    targets: List[dict],
) -> Tuple[Dict[str, dict], Dict[str, dict]]:
    """
    From a list of target structures, parse them into a dictionary with keys
    as target names and data initialization to support future interaction.
    """

    literals: Dict[str, dict] = {}
    patterns: Dict[str, dict] = {}

    for target in targets:
        data: dict = {}
        assert "name" in target
        data["literal"] = target_is_literal(target["name"])
        dest_set = literals if data["literal"] else patterns
        data["data"] = target
        parsed = parse_target(target["name"])
        data["pattern"] = parsed.pattern
        data["keys"] = parsed.keys
        dest_set[target["name"]] = data

    return literals, patterns


MatchData = Dict[str, str]


class TargetMatch(NamedTuple):
    """
    An encapsulation of results when attempting to patch a target name to a
    pattern. If a target was matched and had keyword substitutions, the actual
    values used will be set as match data.
    """

    found: bool
    substitutions: MatchData


def match_target(
    name: str, pattern: re.Pattern, keys: List[str]
) -> TargetMatch:
    """
    From a target name, attempt to match against a pattern and resolve a set
    of key names.
    """

    data: MatchData = defaultdict(str)
    result = pattern.fullmatch(name)

    if result is None:
        return TargetMatch(False, data)

    for idx, key in enumerate(keys):
        data[key] = result.group(1 + idx)

    return TargetMatch(True, data)


def resolve_target_list(target_list: list, match_data: MatchData) -> list:
    """
    Resovle matched-target data into a list form of target data from a
    manifest.
    """

    result: list = []

    for value in target_list:
        if isinstance(value, dict):
            result.append(resolve_target_data(value, match_data))
        elif isinstance(value, list):
            result.append(resolve_target_list(value, match_data))
        elif isinstance(value, str):
            result.append(format_resolve_delims(value, match_data))
        else:
            result.append(value)

    assert len(result) == len(target_list)
    return result


def resolve_dep_data(entry: dict, data: dict) -> dict:
    """
    Implements the business logic for applying match data to manifest entries.
    """

    if "overrides" in entry and "override_path" in entry:
        data = deepcopy(data)
        to_update = advance_dict_by_path(
            entry["override_path"].split("."), data
        )
        if isinstance(to_update, dict):
            merge(
                to_update,
                unflatten_dict(entry["overrides"]),
                expect_overwrite=True,
            )

    return data


def resolve_target_data(target_data: dict, match_data: MatchData) -> dict:
    """Resolve matched-target data into a target's data from a manifest."""

    result: dict = {}

    for key, value in target_data.items():
        if isinstance(value, dict):
            result[key] = resolve_target_data(value, match_data)
        elif isinstance(value, list):
            result[key] = resolve_target_list(value, match_data)
        elif isinstance(value, str):
            result[key] = format_resolve_delims(value, match_data)
        else:
            result[key] = value

    assert len(result.keys()) == len(target_data.keys())
    return result



import logging
from datazen import DEFAULT_TYPE as DEFAULT_TYPE
from typing import Any, Tuple

LOG: Any

def write_dir(directory: str, data: dict, out_type: str = ...) -> None: ...
def str_compile(configs: dict, data_type: str, logger: logging.Logger = ...) -> str: ...
def get_compile_output(entry: dict, default_type: str = ...) -> Tuple[str, str]: ...



# =====================================
# generator=datazen
# version=1.8.0
# hash=d6879e4f3695c6a5760442c423e11c7a
# =====================================

"""
datazen - Package definition for distribution.
"""

# third-party
from vmklib.setup import setup

# internal
from datazen import PKG_NAME, VERSION, DESCRIPTION


author_info = {
    "name": "Vaughn Kottler",
    "email": "vaughnkottler@gmail.com",
    "username": "vkottler",
}
pkg_info = {
    "name": PKG_NAME,
    "slug": PKG_NAME.replace("-", "_"),
    "version": VERSION,
    "description": DESCRIPTION,
    "versions": [
        "3.7",
        "3.8",
        "3.9",
    ],
}
setup(
    pkg_info,
    author_info,
    entry_override="dz",
)



from datazen import ROOT_NAMESPACE as ROOT_NAMESPACE
from datazen.enums import DataType as DataType
from datazen.environment.base import BaseEnvironment as BaseEnvironment
from datazen.load import DEFAULT_LOADS as DEFAULT_LOADS, LoadedFiles as LoadedFiles
from datazen.schemas import inject_custom_schemas as inject_custom_schemas, load_types as load_types, validate as validate
from typing import List

class SchemaEnvironment(BaseEnvironment):
    def load_schema_types(self, sch_loads: LoadedFiles = ..., name: str = ...) -> dict: ...
    def load_schemas(self, require_all: bool = ..., sch_loads: LoadedFiles = ..., sch_types_loads: LoadedFiles = ..., name: str = ..., modify_registry: bool = ...) -> dict: ...
    def enforce_schemas(self, data: dict, require_all: bool = ..., sch_loads: LoadedFiles = ..., sch_types_loads: LoadedFiles = ..., name: str = ...) -> bool: ...
    def add_schema_type_dirs(self, dir_paths: List[str], rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> int: ...
    def add_schema_dirs(self, dir_paths: List[str], rel_path: str = ..., name: str = ..., allow_dup: bool = ...) -> int: ...



from datazen.compile import write_dir as write_dir
from datazen.load import load_dir_only as load_dir_only
from typing import Any

class TaskDataCache:
    data: Any
    cache_dir: Any
    def __init__(self, cache_dir: str): ...
    def load(self, load_dir: str) -> None: ...
    def save(self, out_type: str = ...) -> None: ...
    def clean(self, purge_data: bool = ...) -> None: ...



import logging
from datazen import GLOBAL_KEY as GLOBAL_KEY
from datazen.parsing import set_file_hash as set_file_hash
from datazen.paths import advance_dict_by_path as advance_dict_by_path, get_file_name as get_file_name, get_path_list as get_path_list, walk_with_excludes as walk_with_excludes
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Tuple

LOG: Any

class LoadedFiles(NamedTuple):
    files: Optional[List[str]]
    file_data: Optional[Dict[str, dict]]

DEFAULT_LOADS: Any

def data_added(key: Any, value: Any, data: dict = ...) -> Iterator[dict]: ...
def meld_and_resolve(full_path: str, existing_data: dict, variables: dict, globals_added: bool = ..., expect_overwrite: bool = ..., is_template: bool = ...) -> bool: ...
def load_dir(path: str, existing_data: dict, variables: dict = ..., loads: LoadedFiles = ..., expect_overwrite: bool = ..., are_templates: bool = ..., logger: logging.Logger = ...) -> dict: ...
def load_dir_only(path: str, expect_overwrite: bool = ..., are_templates: bool = ..., logger: logging.Logger = ...) -> dict: ...
def load_files(file_paths: List[str], root: str, meld_data: Tuple[dict, dict, bool], hashes: Dict[str, dict] = ..., expect_overwrite: bool = ..., are_templates: bool = ...) -> List[str]: ...



# =====================================
# generator=datazen
# version=1.8.0
# hash=123bbfe4db73b6f3a12276cd93aa9768
# =====================================

"""
datazen - This package's command-line entry-point (boilerplate).
"""

# built-in
import argparse
import logging
import os
import sys
from typing import List

# internal
from datazen import VERSION, DESCRIPTION
from datazen.app import entry, add_app_args


def main(argv: List[str] = None) -> int:
    """Program entry-point."""

    result = 0

    # fall back on command-line arguments
    command_args = sys.argv
    if argv is not None:
        command_args = argv

    # initialize argument parsing
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="set to increase logging verbosity",
    )
    parser.add_argument(
        "-C",
        "--dir",
        default=os.getcwd(),
        dest="dir",
        help="execute from a specific directory",
    )
    starting_dir = os.getcwd()

    add_app_args(parser)

    # parse arguments and execute the requested command
    try:
        args = parser.parse_args(command_args[1:])
        args.version = VERSION
        args.dir = os.path.abspath(args.dir)

        # initialize logging
        log_level = logging.DEBUG if args.verbose else logging.INFO
        logging.basicConfig(
            level=log_level,
            format="%(name)-36s - %(levelname)-6s - %(message)s",
        )

        # change to the specified directory
        os.chdir(args.dir)

        # run the application
        result = entry(args)
    except SystemExit as exc:
        result = 1
        if exc.code is not None:
            result = exc.code

    # return to starting dir
    os.chdir(starting_dir)

    return result



import logging
from datazen.targets import MatchData as MatchData, match_target as match_target, parse_targets as parse_targets, resolve_target_data as resolve_target_data
from typing import Any, List, Optional

class TargetResolver:
    literals: Any
    patterns: Any
    logger: Any
    def __init__(self, logger: logging.Logger = ...) -> None: ...
    def clear(self) -> None: ...
    def get_target(self, group: str, name: str) -> Optional[dict]: ...
    def register_group(self, name: str, targets: List[dict]) -> None: ...



import logging
from datazen.compile import get_compile_output as get_compile_output, str_compile as str_compile
from datazen.environment.base import TaskResult as TaskResult
from datazen.environment.task import TaskEnvironment as TaskEnvironment
from datazen.parsing import merge_dicts as merge_dicts
from datazen.paths import advance_dict_by_path as advance_dict_by_path
from datazen.targets import resolve_dep_data as resolve_dep_data
from typing import Any, List

LOG: Any

class CompileEnvironment(TaskEnvironment):
    def __init__(self) -> None: ...
    def valid_compile(self, entry: dict, namespace: str, dep_data: dict = ..., deps_changed: List[str] = ..., logger: logging.Logger = ...) -> TaskResult: ...



import jinja2
import logging
from datazen import GLOBAL_KEY as GLOBAL_KEY, to_private as to_private
from datazen.environment.base import TaskResult as TaskResult, dep_slug_unwrap as dep_slug_unwrap
from datazen.environment.task import TaskEnvironment as TaskEnvironment, get_path as get_path
from datazen.fingerprinting import build_fingerprint as build_fingerprint
from datazen.load import data_added as data_added
from datazen.paths import get_file_ext as get_file_ext
from datazen.targets import resolve_dep_data as resolve_dep_data
from typing import Any, Dict, List, Optional

LOG: Any

def render_name_to_key(name: str) -> str: ...
def indent_str(data: str, indent: int, sep: str = ...) -> str: ...
def get_render_str(template: jinja2.Template, name: str, indent: int, data: dict = ..., out_data: dict = ...) -> str: ...
def get_render_children(children: dict, dep_data: dict, default_op: str, indent: int, delimeter: str = ...) -> None: ...

class RenderEnvironment(TaskEnvironment):
    def __init__(self) -> None: ...
    def perform_render(self, template: jinja2.Template, all_templates: Dict[str, jinja2.Template], path: Optional[str], entry: dict, data: dict = ..., logger: logging.Logger = ...) -> TaskResult: ...
    def store_render(self, entry: dict, data: dict) -> None: ...
    def valid_render(self, entry: dict, namespace: str, dep_data: dict = ..., deps_changed: List[str] = ..., logger: logging.Logger = ...) -> TaskResult: ...



from datazen import PKG_NAME as PKG_NAME
from typing import Any, Iterator, List, Tuple

FMT_OPEN: str
FMT_CLOSE: str
EXCLUDES: Any

def get_file_name(full_path: str) -> str: ...
def get_file_ext(full_path: str) -> str: ...
def get_path_list(root_abs_path: str, current_abs_path: str) -> List[str]: ...
def format_resolve_delims(value: str, fmt_data: dict, delim: str = ..., delim_replace: str = ...) -> str: ...
def unflatten_dict(data: dict, delim: str = ...) -> dict: ...
def advance_dict_by_path(path_list: List[str], data: dict) -> dict: ...
def get_package_data(relative_path: str) -> str: ...
def get_package_dir(relative_path: str) -> str: ...
def resolve_dir(data: str, rel_base: str = ...) -> str: ...
def walk_with_excludes(path: str, excludes: List[str] = ...) -> Iterator[Tuple[str, List[str], List[str]]]: ...
